/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex;

import messif.algorithm.AlgorithmMethodException;
import messif.bucket.*;
import messif.bucket.split.SplitPolicy;
import messif.bucket.split.SplitResult;
import messif.data.DataObject;
import messif.distance.DistanceFunc;
import messif.indexing.IndexObject;
import messif.utility.ExtendedPropertiesException;
import mindex.navigation.MIndexInternalCell;
import mindex.navigation.MIndexLeafCell;
import mindex.navigation.MIndexPreciseInternalCell;
import mindex.navigation.VoronoiInternalCell;

import java.io.IOException;
import java.util.*;
import java.util.logging.Level;

/**
 * Created by david on 27/10/16.
 */
public class MIndex extends VoronoiTreeIndex<IndexObject> {

    /** Class serial ID for Java serialization. */
    private static final long serialVersionUID = 1012120L;

    // region ********************     Fields necessary just for the M-Index extension of the Voronoi tree

    /** The bucket dispatcher for buckets in the clusterStructure */
    protected final BucketDispatcher bucketDispatcher;

    /** Class for small clusters */
    private Class<? extends LocalBucket> overfilledBucketClass;

    /** Class for small clusters */
    private Map<String, Object> overfilledBucketParams;

    /** Returns this M-Index bucket dispatcher */
    public BucketDispatcher getBucketDispatcher() {
        return bucketDispatcher;
    }
    // endregion

    /**
     * Creates an M-Index given the config in properties
     * @param origConfiguration config in form of properties
     * @param prefix prefix of the config
     * @throws InstantiationException
     * @throws AlgorithmMethodException
     */
    public MIndex(Properties origConfiguration, String prefix) throws InstantiationException, AlgorithmMethodException {
        this(origConfiguration, prefix, null);
    }

    /**
     * Creates an M-Index given the config in properties
     * @param origConfiguration config in form of properties
     * @param prefix prefix of all
     * @param distanceFunc an instance of distance function that has priority of the text-based config
     * @throws InstantiationException
     * @throws AlgorithmMethodException
     */
    public MIndex(Properties origConfiguration, String prefix, DistanceFunc<DataObject> distanceFunc) throws InstantiationException, AlgorithmMethodException {
        this(new MIndexProperties(origConfiguration, prefix), distanceFunc);
    }

    protected MIndex(MIndexProperties config, DistanceFunc<DataObject> distanceFunc) throws InstantiationException, AlgorithmMethodException {
        super(config, distanceFunc);
        try {
            this.bucketDispatcher = createBucketDispatcher(config);
            // initialize config for overfilled buckets (split is not possible any more)
            this.overfilledBucketClass =  config.getClassProperty(MIndexProperty.OVERFILLED_BUCKET_CLASS.getKey(), false, LocalBucket.class);
            this.overfilledBucketParams = fillBucketProperties(MIndexProperty.PREFIX_OVERFILLED_BUCKET_PARAMS, config);
        } catch (ExtendedPropertiesException | NumberFormatException ex) {
            MetricIndexes.logger.log(Level.SEVERE, ex.toString(), ex);
            throw new InstantiationException(ex.toString());
        }
    }

    /**
     * Create a bucket dispatcher given properties with M-Index config.
     * @param configuration
     * @return
     * @throws InstantiationException
     */
    protected static BucketDispatcher createBucketDispatcher(MIndexProperties configuration) throws InstantiationException {
        Class<? extends LocalBucket> bucketClass = configuration.getClassProperty(MIndexProperty.BUCKET_CLASS.getKey(), true, LocalBucket.class);
        Map<String, Object> bucketClassParams = fillBucketProperties(MIndexProperty.PREFIX_BUCKET_PARAMS, configuration);
        long bucketHardCapacity = configuration.getLongProperty(MIndexProperty.BUCKET_MAX_CAPACITY);
        boolean bucketOccupationAsBytes = configuration.getBoolProperty(MIndexProperty.OCCUPATION_IN_BYTES);
        //long bucketMinOccupation = config.getLongProperty(MIndexProperty.BUCKET_MIN_OCCUPATION);
        return new BucketDispatcher(Integer.MAX_VALUE, bucketHardCapacity, bucketHardCapacity, 0L, bucketOccupationAsBytes, bucketClass, bucketClassParams);
    }

    /**
     * THis method fills and returns properties for bucket creating from this config.
     *   String THESE_PROPERTIES is substituted by the original config properties file.
     * @param paramsPrefix prefix of the bucket properties, e.g. "bucket.params."
     * @param configuration the whole M-Index config
     * @return new config properties for bucket creation
     */
    public static Map<String, Object> fillBucketProperties(String paramsPrefix, Properties configuration) {
        Map<String, Object> retval = new HashMap<>();
        for (String key : configuration.stringPropertyNames()) {
            if (key.startsWith(paramsPrefix)) {
                if ("THESE_PROPERTIES".equals(configuration.getProperty(key))) {
                    retval.put(key.substring(paramsPrefix.length()), configuration);
                } else {
                    retval.put(key.substring(paramsPrefix.length()), configuration.getProperty(key));
                }
            }
        }
        return retval;
    }

    protected VoronoiInternalCell initVoronoiCellTree() throws AlgorithmMethodException {
        return isPreciseSearch() ? new MIndexPreciseInternalCell(this, null, new short[0])
                : new MIndexInternalCell(this, null, new short[0]);
    }

    @Override
    public IndexObject initObject(DataObject object) {
        return ppCalculator.initMIndexObject(object);
    }

//    @Override
//    public IndexObject initObject(DataObject object, boolean setObjPivotDistance, boolean allObjPivotDistances) {
//        return ppCalculator.initMIndexObject(object, setObjPivotDistance, allObjPivotDistances);
//    }
//
    @Override
    public Collection<IndexObject> initObjects(DataObject [] objects) {
        return ppCalculator.initMIndexObjects(objects);
    }

    @Override
    public short[] getPPP(IndexObject object) {
        return object.getIndexField(MIndexKeys.PPP, short[].class);
    }

    @Override
    public DataObject getDataObject(IndexObject indexObject) {
        return indexObject.getDataObject();
    }


    // region *********************  Methods taking care of the bucket storage **********************

    /**
     * Return the number of objects stored by this M-Index.
     * @return the number of objects stored by this M-Index.
     */
    public int getObjectCount() {
        return (bucketDispatcher != null) ? bucketDispatcher.getObjectCount() : voronoiCellTree.getObjectCount();
    }

    /**
     * The given bucket is removed from management of this M-Index (from respective bucket dispatcher).
     * @param bucket bucket to be removed
     * @param ignoreNonexisting
     * @throws AlgorithmMethodException is this bucket is not managed by this M-Index
     */
    public void removeBucket(Bucket bucket, boolean ignoreNonexisting) throws AlgorithmMethodException {
        try {
            bucketDispatcher.removeBucket(bucket.getBucketID());
        } catch (NoSuchElementException e) {
            if (! ignoreNonexisting) {
                throw new AlgorithmMethodException("Given bucket is not managed by this M-Index and cannot be removed: " + e.getLocalizedMessage());
            }
        }
    }

    /**
     * Creates a bucket for given leaf cell.
     */
    public Bucket createBucket(MIndexLeafCell leaf) throws BucketStorageException, IOException, AlgorithmMethodException {
        return bucketDispatcher.createBucket();
    }

    /**
     * This method is overridden in the DistributedMIndex
     * @param bucket
     * @param policy
     * @param whoStays
     * @return
     * @throws BucketStorageException
     */
    public SplitResult splitLeafBucket(Bucket bucket, final SplitPolicy policy, int whoStays) throws BucketStorageException {
        final List<Bucket> newBuckets = new ArrayList<>(getNumberOfPivots());
        final int objectsMoved = bucket.split(policy, newBuckets, bucketDispatcher, whoStays);

        return new SplitResult() {
            @Override
            public List<? extends Bucket> getBuckets(BucketDispatcher bucketDisp) throws CapacityFullException {
                return newBuckets;
            }
            @Override
            public SplitPolicy getSplitPolicy() {
                return policy;
            }
            @Override
            public int getObjectsMoved() {
                return objectsMoved;
            }
        };
    }

    /**
     * If a special bucket for overfilled cells is specified than the data is reorganized in this new bucket.
     *  If the data cannot be reorganized (the special bucket type is not set or it has already been reorganized)
     *  then NULL is returned.
     *
     * @param origBucket bucket to be reorganized
     * @return new created bucket or <code>null</code>, if the data cannot be reorganized
     *      (the special bucket type is not set or it has already been reorganized)
     * @throws AlgorithmMethodException if anything goes wrong during new bucket creation
     */
    public Bucket reorganizeOvefilled(Bucket origBucket) throws AlgorithmMethodException {
        try {
            LocalBucket newBucket;
            if (! isSpecialOverfilledBuckets() || overfilledBucketClass.isInstance(origBucket)) {
                newBucket = bucketDispatcher.createBucket(Long.MAX_VALUE, Long.MAX_VALUE, bucketDispatcher.getBucketLowOccupation());
            } else {
                newBucket = bucketDispatcher.createBucket(overfilledBucketClass, overfilledBucketParams, Long.MAX_VALUE, Long.MAX_VALUE, bucketDispatcher.getBucketLowOccupation());
            }
            try {
                newBucket.addObjects(origBucket.getAllObjects());
            } catch (NoSuchElementException e) {
                MetricIndexes.logger.warning("reorganization of overfilled bucket initialized but the bucket was actually empty: ID=" + origBucket.getBucketID());
            }
            bucketDispatcher.removeBucket(origBucket.getBucketID());

            return newBucket;
        } catch (BucketStorageException ex) {
            MetricIndexes.logger.log(Level.SEVERE, ex.toString(), ex);
            throw new AlgorithmMethodException(ex);
        }
    }
    // endregion


    @Override
    @SuppressWarnings({"FinalizeDeclaration", "FinalizeCalledExplicitly"})
    public void finalize() throws Throwable {
        bucketDispatcher.finalize();
        super.finalize();
    }

    public void destroy() throws Throwable {
        bucketDispatcher.destroy();
    }

    /**
     * Prints detailed information about this M-Index - including the dynamic cluster tree
     *  and numbers of objects stored in individual clusters.
     * WARNING: opens all disk buckets (files)
     * @return string representation of this M-Index
     */
    public String toRichString() {
        StringBuilder strBuf = new StringBuilder(super.toRichString());
        strBuf.append("\nPhysical buckets:\n");
        int sumOfObjects = 0;
        for (LocalBucket bucket : bucketDispatcher.getAllBuckets()) {
            strBuf.append("\tbucket ").append(bucket.getBucketID()).append(": storing ").append(bucket.getOccupation()).append("\n");
            sumOfObjects += bucket.getOccupation();
        }
        strBuf.append("\nThe bucket dispatcher stores: ").append(sumOfObjects).append(" objects (bytes)\n");
        return strBuf.toString();
    }

    /**
     * Prints info about this M-Index.
     * @return string representation of this M-Index
     */
    @Override
    public String toString() {
        StringBuilder strBuf = new StringBuilder(super.toString());
        // storage information
        strBuf.append("Storage: ").append(bucketDispatcher).append("\n");
        return strBuf.toString();
    }

}
