/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex;

import messif.data.DataObject;
import messif.data.util.DataObjectList;
import messif.data.util.StreamDataObjectIterator;
import messif.distance.DistanceFunc;
import messif.indexing.IndexObject;
import messif.indexing.PrecomputedDistancesFixedArrayFilter;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.util.*;
import java.util.logging.Level;
import java.util.stream.Stream;

/**
 * This is an abstract class carries part of the M-Index config that
 *   deals with pivots and calculation of distances between object and pivots.
 * It can especially create the M-Index pivot permutation (M-Index key).
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class PPPCalculator implements Serializable {

    /** Class serial ID for serialization */
    private static final long serialVersionUID = 112080L;
    
    /** The whole M-Index config is kept in the extended properties file */
    protected final MIndexProperties configuration;
    
    /** The set of pivots are fixed during the whole life of the index - can be null. */
    protected final DataObject[] pivots;

    /** Distance function used by this M-Index. */
    protected final DistanceFunc<DataObject> distanceFunc;
    
    // ************************   Transient fields (depending on the actual config)  ***************** //

    /** Maximal M-Index partitioning level */
    protected transient int maxLevel;
    
    /** Size of the pivots set */
    protected transient long maxNumberOfClusters;
    
    /** Number of pivots - it can depend on several aspects. */
    protected transient int numberOfPivots;
    
    /** Matrix of mutual pivot distances */
    protected transient float[][] pivotDistancesMatrix = null;

    
    // ****************************       Initialization      ************************************* //
    
    /**
     * Creates the PP calculator given M-Index properties
     * @param mIndexProperties properties with M-Index config
     * @throws InstantiationException if this part of M-Index cannot be initialized using this properties
     */
    public PPPCalculator(MIndexProperties mIndexProperties, DistanceFunc<DataObject> distanceFunc) throws InstantiationException {
        this.configuration = mIndexProperties;
        // handle the pivots - either read them from the file or only the number is specified
        this.pivots = initPivots(configuration);
        this.distanceFunc = distanceFunc == null ?
                mIndexProperties.getObjectProperty(MIndexProperty.DISTANCE_FUNC, DistanceFunc.class) : distanceFunc;
        setPrecomputedFields();
    }

    /** 
     * Given maximal specified number of pivots and config property file, this method
     *  returns the prepared list of pivot objects.
     * @param configuration M-Index config
     * @return the prepared list of pivot objects or <b>null</b> if file does not exist or is empty
     */
    public static DataObject[] initPivots(MIndexProperties configuration) throws InstantiationException{
        try {
            int nMaxPivots = configuration.getIntProperty(MIndexProperty.NUMBER_OF_PIVOTS);
            if (! new File(configuration.getProperty(MIndexProperty.PIVOT_FILE)).canRead()) {
                throw new InstantiationException("cannot find and read pivot file: " + configuration.getProperty(MIndexProperty.PIVOT_FILE));
            }
            // read the pivots from file
            StreamDataObjectIterator iterator = new StreamDataObjectIterator(configuration.getProperty(MIndexProperty.PIVOT_FILE));
            DataObjectList pivotsList = (nMaxPivots >= 0) ?
                    new DataObjectList(iterator, nMaxPivots) : new DataObjectList(iterator);

            if (pivotsList.size() <= 0) {
                throw new InstantiationException("the pivot file is empty or contains no valid data objects: "
                        + configuration.getProperty(MIndexProperty.PIVOT_FILE));
            }
            return pivotsList.toArray(new DataObject[pivotsList.size()]);
        } catch (IOException io) {
            throw new InstantiationException(io.toString());
        }
    }
        
    /**
     * Deserialization method - update the precomputed fields.
     * @param in stream
     * @throws java.io.IOException
     * @throws java.lang.ClassNotFoundException
     */
    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        try {
            if (configuration != null) {
                setPrecomputedFields();
            }
        } catch (InstantiationException ex) {
            throw new IOException(ex);
        }
    }
    
    /**
     * This method sets all the transient fields - it is called from the constructor 
     *  and when the object is being deserialized.
     */
    private void setPrecomputedFields() throws InstantiationException {
        this.maxLevel = configuration.getIntProperty(MIndexProperty.MAX_LEVEL);
        this.numberOfPivots = pivots.length;
        // if the maximal number of clusters should be larger than Long.MAX then print warning (not fatal for PPP-Codes)
        if (Math.pow(numberOfPivots, maxLevel) > Long.MAX_VALUE) {
            MetricIndexes.logger.log(Level.SEVERE, "having {0} pivots and max level {1}, the theoretical number of " +
                            "clusters exceeds the Long.MAX (not fatal for PPP-Codes)", new Object [] { getNumberOfPivots(), getMaxLevel()});
            this.maxNumberOfClusters = Long.MAX_VALUE;
        } else {
            this.maxNumberOfClusters = (long) Math.pow(numberOfPivots, maxLevel);
        }
    }

    private void initPivotDistMatrix() {
        // init the matrix of pivot-query distances
        this.pivotDistancesMatrix = new float[numberOfPivots][numberOfPivots];
        for (int i = 0; i < numberOfPivots; i++) {
            for (int j = 0; j < i; j++) {
                this.pivotDistancesMatrix[i][j] = distanceFunc.getDistance(pivots[i], pivots[j]);
                this.pivotDistancesMatrix[j][i] = this.pivotDistancesMatrix[i][j];
            }
            this.pivotDistancesMatrix[i][i] = 0f;
        }        
    }
    
    // ****************************       Getters      ************************************* //
    
    /** 
     * Getter for individual pivots.
     * @param pivotNumber pivot number to get [0-(n-1)]
     * @return a pivot given a number
     */
    public DataObject getPivot(int pivotNumber) {
        return pivots[pivotNumber];
    }

    /**
     * Returns distance between given two pivots, if the pivots are set; it is 
     *  precomputed in a pivot x pivot matrix.
     * @param pivot1 the first pivot index
     * @param pivot2 the second pivot index
     * @return distance between the pivots or -1 (if pivots not set)
     */
    public float getPivotsDistance(int pivot1, int pivot2) {
        if (pivotDistancesMatrix == null) {
            initPivotDistMatrix();
        }
        return this.pivotDistancesMatrix[pivot1][pivot2];
    }
    
    
    /** Number of pivots (it IS always known). */
    public int getNumberOfPivots() {
        return numberOfPivots;
    }

    /** Returns the maximal number of clusters created by this M-Index: $(number-of-pivots)^level$ */
    public long getMaxNumberOfClusters() {
        return this.maxNumberOfClusters;
    }
    
    public boolean isUsePivotFiltering() {
        return configuration.getBoolProperty(MIndexProperty.PIVOT_FILTERING);
    }

    public boolean isCheckDuplicateDC() {
        return configuration.getBoolProperty(MIndexProperty.CHECK_DUPLICATE_DC);
    }
    
//    public boolean isUseExistingPP() {
//        return config.getBoolProperty(MIndexProperty.USE_EXISTING_PP);
//    }

    public boolean isPreciseSearch() {
        return configuration.getBoolProperty(MIndexProperty.PRECISE_SEARCH);
    }

    /** Return the maximal level of the M-Index. */
    public int getMaxLevel() {
        return maxLevel;
    }

    /** Returns the distance function of this M-Index. */
    public DistanceFunc<DataObject> getDistanceFunc() {
        return distanceFunc;
    }

    // ***************       Init object - calculate obj-pivot distance + PP  *********************  //
    
    /**
     * Initialize the {@link IndexObject} to be stored in M-Index: calculate all distances, sort them, set
     * pivot permutation (and set distance filter, if the filter should be stored by this M-Index).
     *
     * @param object to initialize
     * @return the initiated index object
     */
    public IndexObject initMIndexObject(DataObject object) {
        return initMIndexObject(object, isPreciseSearch(), isUsePivotFiltering());
    }

    /**
     * Return a PPP for an object. If it is an {@link IndexObject} of M-Index, we don't have to compute anything, just
     *  return the precomputed PPP; otherwise, use {@link #initMIndexObject(DataObject, boolean, boolean)} and return the PPP.
     */
    public short [] getPPP(DataObject object) {
        short [] retVal = MIndexKeys.readMIndexPP(object);
        if (retVal != null) {
            return retVal;
        }
        return sortDistances(getObjectPivotDistances(object), maxLevel);
    }

    /**
     * Initialize the object in M-Index: calculate all distances, sort them, set pivot simple permutation (if precise search
     *  is not ON) or full M-Index key (if precise search is ON) and set filter (if filtering required).
     * After calling this method, the {@code object} has only one MIndexPP and that is the one created here.
     * @param object object to initialize
     * @param allObjPivotDistances if the distances should be returned, independent of the pivot filtering usage
     * @return the object-pivot distances, or null if the {@code getDistances} is false
     */
    public IndexObject initMIndexObject(DataObject object, boolean setObjPivotDistance, boolean allObjPivotDistances) {
        final float[] objectPivotDistances = getObjectPivotDistances(object);
        short[] ppp = sortDistances(objectPivotDistances, maxLevel);

        IndexObject retVal = (object instanceof IndexObject) ? (IndexObject) object : new IndexObject(object);
        retVal.setIndexField(MIndexKeys.PPP, ppp);
        if (setObjPivotDistance) {
            retVal.setIndexField(MIndexKeys.OBJ_PIVOT_DISTANCE, objectPivotDistances[ppp[0]]);
        }
        if (allObjPivotDistances) {
            retVal.setIndexField(MIndexKeys.OBJ_PIVOT_DISTANCES, objectPivotDistances);
        }
        if (isUsePivotFiltering()) {
            PrecomputedDistancesFixedArrayFilter filter = new PrecomputedDistancesFixedArrayFilter(objectPivotDistances);
            retVal.setIndexField(PrecomputedDistancesFixedArrayFilter.FILTER_STRING, filter);
        }

        if (isCheckDuplicateDC() && ! retVal.containsIndexField(DuplicateRefineFilter.FILTER_STRING)) {
            retVal.setIndexField(DuplicateRefineFilter.FILTER_STRING, new DuplicateRefineFilter(object.getID()));
        }

        return retVal;
    }

    /**
     * Initializes the PP of a collection of objects. If the thread pool is created, it is done in parallel.
     *
     * @param objects the collection of objects to initialize the PP for
     * @return collection of initialized objects - it may differ from the passed one
     */
    public Collection<IndexObject> initMIndexObjects(DataObject[] objects) {
        Stream<DataObject> stream = Arrays.stream(objects);
        List<IndexObject> retVal;
        if (configuration.getBoolProperty(MIndexProperty.PARALLEL_INIT)) {
            stream = stream.parallel();
            retVal = Collections.synchronizedList(new ArrayList<>(objects.length));
        } else {
            retVal = new ArrayList<>(objects.length);
        }

        stream.forEach((object) -> retVal.add(initMIndexObject(object)));
        return retVal;
    }


    //  ******************    Internal methods to be exploited by the abstract methods    ***************  //
    
    /**
     * Calculate distances between given object and all pivots.
     * @param object data object to calculate the distances for
     * @return array of object-pivot distances
     */
    public float [] getObjectPivotDistances(DataObject object) {
        float[] distances = new float[numberOfPivots];
        for (int distanceIndex = 0; distanceIndex < numberOfPivots; distanceIndex++) {
            distances[distanceIndex] = distanceFunc.getDistance(pivots[distanceIndex], object);
        }
        return distances;
    }

    /**
     * Sort the current array creating the array of indexes and return it.
     * @param distances the distance array to be sorted
     * @param maxLevel keep only a given prefix of the sorting indexes
     * @return array of sorting indexes (pivot permutation of given length)
     */
    public static short[] sortDistances(float [] distances, int maxLevel) {
        DistanceIndexPair[] actualValues = new DistanceIndexPair[distances.length];
        for (short i = 0; i < distances.length; i++) {
            actualValues[i] = new DistanceIndexPair(distances[i], i);
        }
        Arrays.sort(actualValues);

        short [] sortingIndexes = new short[maxLevel];
        for (int i = 0; i < maxLevel; i ++ ) {
            sortingIndexes[i] = actualValues[i].index;
        }
        return sortingIndexes;
    }
    
    /**
     * Auxiliary class for efficient sorting of the distances and determining the sorting indexes.
     */
    protected static class DistanceIndexPair implements Comparable {

        protected final float distance;
        
        protected final short index;

        public DistanceIndexPair(float distance, short index) {
            this.distance = distance;
            this.index = index;
        }                
        
        @Override
        public int compareTo(Object o) {
            return Float.compare(distance, ((DistanceIndexPair)o ).distance);
        }
    }
    
    /**
     * This method returns the number of cluster corresponding to this PP an level; 
     *  it also depends on the maxlevel (size of PP).
     * @param pivotPermutation PP to create the long Key value from
     * @param level of the desired long key (1-based)
     * @return M-Index number of corresponding cluster
     */
    protected long computeClusterNumber(short [] pivotPermutation, int level) {
        long retVal = 0L;
        long constant = 1L;
        for (int i = maxLevel - 1; i >= 0; i--) {
            if (i < level && i < pivotPermutation.length) {
                retVal += pivotPermutation[i] * constant;
            }
            constant *= numberOfPivots;
        }
        return retVal;
    }

    @Override
    public String toString() {
        StringBuilder strBuf = new StringBuilder();
        strBuf.append("distance func: ").append(distanceFunc.toString()).append("\n");
        strBuf.append("max # of clusters = ").append(maxNumberOfClusters).append("; # of pivots = ").append(numberOfPivots).append(": \n");
        for (int i = 0; i < numberOfPivots; i++) {
            strBuf.append(getPivot(i).getID()).append(", ");
        }
        return strBuf.toString();
    }    
    
}
