/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex;


import messif.algorithm.AlgorithmMethodException;
import messif.data.DataObject;
import messif.distance.DistanceFunc;
import messif.operation.params.ApproximateSearch;
import mindex.navigation.VoronoiInternalCell;
import mindex.navigation.VoronoiLeafCell;

/**
 * This class carries the core functionality of the M-Index or a single PPP-Codes tree
 * <ul>
 *  <li> having an object, it calculates its M-Index keys (PPP + distance)</li>
 *  <li> it maintains the dynamic Voronoi tree</li>
 *  <li> it has the bucket dispatcher</li>
 * </ul>
 * 
 * It also carries the parameters of the Voronoi tree - maximum level, (minimum level), and pivots
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public abstract class VoronoiTreeIndex<O> implements java.io.Serializable, IndexObjectInitializer<O> {

    // region **************************    Fields    *****************************

    /** Class serial ID for Java serialization. */
    private static final long serialVersionUID = 11211L;
    
    /** The whole M-Index config is kept in the extended properties file */
    protected final MIndexProperties configuration;
    
    /** The M-Index structure - dynamic levels */
    protected final VoronoiInternalCell voronoiCellTree;

    /** This field encapsulates the M-Index pivot and calculation of PP */
    protected final PPPCalculator ppCalculator;
    
    /** Flag that says whether the dynamic cell tree was modified and should be serialized */
    protected transient volatile boolean dynamicTreeModified = false;
    // endregion


    // region  ************************************   Getters     ************************

    /** Returns the M-Index Pivot Permutation (PP) calculator */
    public PPPCalculator getPPCalculator() {
        return ppCalculator;
    }


    /** Returns the dynamic M-Index tree root */
    public VoronoiInternalCell getVoronoiCellTree() {
        return voronoiCellTree;
    }
    
    /** Number of pivots (it IS always known). */
    public int getNumberOfPivots() {
        return ppCalculator.getNumberOfPivots();
    }

    /** Returns the distance function of this M-Index. */
    public DistanceFunc<DataObject> getDistanceFunc() {
        return ppCalculator.getDistanceFunc();
    }

    /**
     * Returns distance between given two pivots, if the pivots are set; it is 
     *  precomputed in a pivot x pivot matrix.
     * @param pivot1 the first pivot index
     * @param pivot2 the second pivot index
     * @return distance between the pivots or -1 (if pivots not set)
     */
    public float getPivotsDistance(int pivot1, int pivot2) {
        return ppCalculator.getPivotsDistance(pivot1, pivot2);
    }
    
    /** Return the maximal level of the M-Index. */
    public int getMaxLevel() {
        return configuration.getIntProperty(MIndexProperty.MAX_LEVEL);
    }

    /** Return the minimal level of the M-Index. */
    public int getMinLevel() {
        return configuration.getIntProperty(MIndexProperty.MIN_LEVEL);
    }
    
    /** Returns the minimal occupation of a self-standing disk bucket; if > 0 then multi-buckets can be created. */
    public long getBucketMinOccupation() {
        return configuration.getLongProperty(MIndexProperty.BUCKET_MIN_OCCUPATION);
    }

    public boolean isSpecialOverfilledBuckets() {
        return configuration.getBoolProperty(MIndexProperty.BUCKET_OVERFILLED_SPECIAL);
    }
        
    public boolean isRemoveDuplicates() {
        return configuration.getBoolProperty(MIndexProperty.REMOVE_DUPLICATES);
    }
        
    public int getApproxDefaultParam() {
        return configuration.getIntProperty(MIndexProperty.APPROX_DEFAULT);
    }

    public ApproximateSearch.Type getApproxDefaultType() {
        return ApproximateSearch.Type.valueOf(configuration.getProperty(MIndexProperty.APPROX_DEFAULT_TYPE));
    }   

    public boolean isForceDefaultApprox() {
        return configuration.getBoolProperty(MIndexProperty.APPROX_FORCE_DEFAULT);
    }   

    public boolean isApproxProcessWholeMultiBucket() {
        return configuration.getBoolProperty(MIndexProperty.APPROX_PROCESS_WHOLE_MULTIBUCKET);
    }

    public boolean isApproxCheckKeyInterval() {
        return configuration.getBoolProperty(MIndexProperty.APPROX_CHECK_KEYINTERVAL);
    }
    
    public boolean isMultiBuckets() {
        return configuration.getLongProperty(MIndexProperty.BUCKET_MIN_OCCUPATION) > 0L;
        //return getBucketDispatcher().getBucketLowOccupation() > 0;
    }
        
    public boolean isAgile() {
        return isMultiBuckets() && configuration.getBoolProperty(MIndexProperty.AGILE_SPLIT);
    }
    
    public boolean isPreciseSearch() {
        return ppCalculator.isPreciseSearch();
    }
    
    public int getBucketCapacity() {
        return (int) configuration.getLongProperty(MIndexProperty.BUCKET_MAX_CAPACITY);
    }
        
    public boolean isBucketOccupationInBytes() {
        return configuration.getBoolProperty(MIndexProperty.OCCUPATION_IN_BYTES);
    }
    
    public boolean isCheckDuplicateDC() {
        return ppCalculator.isCheckDuplicateDC();
    }
    // endregion

    
    // region **************************************    Initialization   ************************************* //

    /**
     * Creates new instance of M-Index having practically all components already initialized.
     * @param config properties specifying M-Index config
     * @throws java.lang.InstantiationException
     */
    protected VoronoiTreeIndex(MIndexProperties config, DistanceFunc<DataObject> distanceFunc) throws InstantiationException, AlgorithmMethodException {
        this.configuration = config;
        this.ppCalculator = new PPPCalculator(config, distanceFunc);
        this.voronoiCellTree = initVoronoiCellTree();
    }
    
    protected abstract VoronoiInternalCell initVoronoiCellTree() throws AlgorithmMethodException;
    // endregion


    // region ******************  Methods taking care of the dynamic levels of the Voronoi tree  **********************

    /** Set the new value of the flag {@link VoronoiTreeIndex#dynamicTreeModified} */
    public void setDynamicTreeModified(boolean dynamicTreeModified) {
        this.dynamicTreeModified = dynamicTreeModified;
    }

    /** Returns the actual value of {@link VoronoiTreeIndex#dynamicTreeModified} */
    public boolean isDynamicTreeModified() {
        return dynamicTreeModified;
    }
    
    /** 
     * Clears (removes) all the data stored in this M-Index. The structure of the 
     *  dynamic cell tree remains untouched.
     * @throws AlgorithmMethodException
     */
    public void clearData() throws AlgorithmMethodException {
        voronoiCellTree.clearData();
    }
    
    /**
     * THis method initializes the object's PPP, then traverses the dynamic PPP-Tree and returns 
     *  the dynamic hash number for given PPP and actual level.
     * @param object data object to find
     * @return the dynamic M-Index key (long) for given object (PP) and level of that specific PP in current PPP-Tree
     * @throws AlgorithmMethodException 
     */
    public long getDynamicKey(DataObject object) throws AlgorithmMethodException {
        short[] objectPPP = ppCalculator.getPPP(object);
        VoronoiLeafCell responsibleLeaf = voronoiCellTree.getResponsibleLeaf(objectPPP, false);
        return ppCalculator.computeClusterNumber(objectPPP, responsibleLeaf.getLevel());
    }
    // endregion

    
    // *********************************************   Auxilionary methods   ********************************* //

    /**
     * Prints detailed information about this M-Index - including the dynamic cluster tree
     *  and numbers of objects stored in individual clusters.
     * WARNING: opens all disk buckets (files)
     * @return string representation of this M-Index
     */
    public String toRichString() {
        StringBuilder strBuf = new StringBuilder(toString());
        // print the dynamic levels including number of objects stored
        strBuf.append(voronoiCellTree.toString());
        return strBuf.toString();
    }

    /** 
     * Prints info about this M-Index.
     * @return string representation of this M-Index
     */
    @Override
    public String toString() {
        StringBuilder strBuf = new StringBuilder();
        for (String property : configuration.stringPropertyNames()) {
            strBuf.append("\t").append(property).append(" = ").append(configuration.getProperty(property)).append('\n');
        }
        // pivot settings information
        strBuf.append(ppCalculator.toString());
        
        // storage information
        strBuf.append("use the ").append(isAgile() ? "AGILE" : "LAZY").append(" approach for partitioning\n");
        strBuf.append("\nStoring: ").append(voronoiCellTree.getObjectCount()).append(" objects\n");
        return strBuf.toString();
    }
}
