/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.algorithm;


import messif.algorithm.Algorithm;
import messif.algorithm.AlgorithmMethodException;
import messif.data.DataObject;
import messif.distance.DistanceFunc;
import messif.operation.info.ObjectCountOperation;
import messif.operation.info.PrintAllObjectsOperation;
import messif.operation.info.answer.ObjectCountAnswer;
import messif.operation.info.answer.PrintAllObjectsAnswer;
import messif.operation.info.processor.PrintAllObjectsProcessorBase;
import messif.operation.processing.ProcessorConfiguration;
import messif.operation.processing.impl.OneStepProcessor;
import messif.operation.search.*;
import messif.operation.answer.RankedAnswer;
import messif.utility.ExtendedProperties;
import mindex.MIndex;
import mindex.processor.ApproxQueryProcessor;
import mindex.processor.KNNOperationProgressiveProcessor;

import java.io.IOException;

/**
 * This is the main M-Index centralized algorithm.
 *
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class MIndexAlgorithm extends VoronoiAlgorithm {

    /** Class serial id for serialization. */
    private static final long serialVersionUID = 20203L;

    /**
     * Creates the M-Index algorithm given the file name file name containing M-Index properties for new M-Index.
     *
     * @param properties text file with M-Index properties settings
     * @param prefix prefix (e.g. "mindex.") for items in property file that should be considered (the rest is ignored)
     * @throws java.lang.InstantiationException if the file does not contain what it should contain
     */
    @Algorithm.AlgorithmConstructor(description = "Constructor from a created properties object", arguments = {"M-Index properties", "prefix"})
    public MIndexAlgorithm(ExtendedProperties properties, String prefix) throws InstantiationException, AlgorithmMethodException {
        this("M-Index", properties, prefix);
    }

    /**
     * Creates the M-Index algorithm given the file name file name containing M-Index properties for new M-Index.
     *
     * @param name algorithm name
     * @param properties text file with M-Index properties settings
     * @param prefix prefix (e.g. "mindex.") for items in property file that should be considered (the rest is ignored)
     * @throws java.lang.InstantiationException if the file does not contain what it should contain
     */
    @Algorithm.AlgorithmConstructor(description = "Constructor from a created properties object", arguments = {"Algorithm name", "M-Index properties", "prefix"})
    public MIndexAlgorithm(String name, ExtendedProperties properties, String prefix) throws InstantiationException, AlgorithmMethodException {
        this(name, properties, prefix, null);
    }

    /**
     * Creates the M-Index algorithm given the file name file name containing M-Index properties for new M-Index.
     *
     * @param name algorithm name 
     * @param properties text file with M-Index properties settings
     * @param prefix prefix (e.g. "mindex.") for items in property file that should be considered (the rest is ignored)
     * @param distanceFunc an instance of distance function to be used instead of the function specified by properties
     * @throws java.lang.InstantiationException if the file does not contain what it should contain
     */
    @Algorithm.AlgorithmConstructor(description = "Constructor from a created properties object", arguments = {"Algorithm name", "M-Index properties", "prefix"})
    public MIndexAlgorithm(String name, ExtendedProperties properties, String prefix, DistanceFunc<DataObject> distanceFunc) throws InstantiationException, AlgorithmMethodException {
        this(name, new MIndex(properties, prefix, distanceFunc), properties);
    }

    /**
     * Creates the algorithm given name and already created MIndex config instance.
     * @param name alg. name
     * @param mIndex created MIndex config object
     */
    protected MIndexAlgorithm(String name, MIndex mIndex, ExtendedProperties algConfiguration) {
        super(name, mIndex, algConfiguration);
    }

    protected void registerProcessors() {
        super.registerProcessors();
        this.handler.register(ProcessorConfiguration.create(ObjectCountOperation.class, ObjectCountAnswer.class)
                .withProcessor(ObjectCountProcessor::new)
                .executeSequential().build());
        this.handler.register(ProcessorConfiguration.create(PrintAllObjectsOperation.class, PrintAllObjectsAnswer.class)
                .withProcessor((operation) -> {
                    try {
                        return new PrintAllObjectsProcessor(operation);
                    } catch (IOException e) {
                        throw new RuntimeException(e);
                    }
                }).build());
        this.handler.register(ProcessorConfiguration.create(KNNOperationApprox.class, RankedAnswer.class)
                .withProcessor((operation) -> new ApproxQueryProcessor(operation, operation, getmIndex()))
                .executeParallel(threadPool).build());
        this.handler.register(ProcessorConfiguration.create(RangeOperationApprox.class, RankedAnswer.class)
                .withProcessor((operation) -> new ApproxQueryProcessor(operation, operation, getmIndex()))
                .executeParallel(threadPool).build());
        this.handler.register(ProcessorConfiguration.create(KNNRangeOperationApprox.class, RankedAnswer.class)
                .withProcessor((operation) -> new ApproxQueryProcessor(operation, operation, getmIndex()))
                .executeParallel(threadPool).build());
        this.handler.register(ProcessorConfiguration.create(KNNOperation.class, RankedAnswer.class)
                .forTypeStrings(KNNOperation.TYPE_STRING, KNNRangeOperation.TYPE_STRING)
                .withProcessor((operation) -> new KNNOperationProgressiveProcessor(operation, getmIndex()))
                .executeParallel(threadPool).build());
    }

    /** Clean up everything */
    @Override
    @SuppressWarnings({"FinalizeDeclaration", "FinalizeCalledExplicitly"})
    public void finalize() throws Throwable {
        if (mIndex != null) {
            getmIndex().finalize();
        }
        super.finalize();
    }

    /** Called before deleting */
    @Override
    public void destroy() throws Throwable {
        super.destroy();
        if (mIndex != null) {
            getmIndex().destroy();
        }
    }


    // ************************      Getters and setters      ************************* //

    /**
     * Sets the period of the automatic temporary-closeable checking of the buckets.
     * @param period the new checking period in milliseconds;
     *          zero value means disable the checking
     */
    public void autoCloseBuckets(long period) {
        getmIndex().getBucketDispatcher().setTemporaryClosePeriod(period);
    }

    /** Getter for internal use within the M-Index (should be removed in future)
     * @return  */
    public MIndex getmIndex() {
        return (MIndex) mIndex;
    }


    // *************************     Operation processing     *************************** //    

    protected class ObjectCountProcessor extends OneStepProcessor<ObjectCountOperation, ObjectCountAnswer> {

        final ObjectCountOperation.Helper helper;

        public ObjectCountProcessor(ObjectCountOperation operation) {
            super(operation);
            helper = new ObjectCountOperation.Helper();
        }

        @Override
        protected void process() throws Exception {
            helper.evaluate(getmIndex().getBucketDispatcher());
        }

        @Override
        public ObjectCountAnswer finish() {
            return helper.getAnswer();
        }
    }


    protected class PrintAllObjectsProcessor extends PrintAllObjectsProcessorBase {

        /**
         * Creates this processor given the operation to be processed.
         *
         * @param operation operation to be executed
         */
        public PrintAllObjectsProcessor(PrintAllObjectsOperation operation) throws IOException {
            super(operation);
        }

        @Override
        public boolean processStep() throws Exception {
            super.evaluate(MIndexAlgorithm.this.getmIndex().getBucketDispatcher());
            return false;
        }
    }

}
