/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.algorithm;


import messif.algorithm.Algorithm;
import messif.algorithm.AlgorithmMethodException;
import messif.algorithm.impl.MultipleOverlaysAlgorithm;
import messif.data.util.RankedSortedCollection;
import messif.indexing.IndexObject;
import messif.operation.OperationBuilder;
import messif.operation.helpers.AnswerProvider;
import messif.operation.params.ApproximateSearch;
import messif.operation.params.QueryObject;
import messif.operation.processing.ProcessorConfiguration;
import messif.operation.processing.impl.AlgorithmsOperationProcessor;
import messif.operation.search.*;
import messif.operation.answer.RankedAnswer;
import mindex.DuplicateRefineFilter;
import mindex.PPPCalculator;
import mindex.VoronoiTreeIndex;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Indexing algorithm that processes operation by passing them to encapsulated collection
 * of M-Index algorithms.
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class MultipleMIndexAlgorithm extends MultipleOverlaysAlgorithm {
    
    /** Class id for serialization. */
    private static final long serialVersionUID = 201002L;

    /** Internal list of algorithms (the same as encapsulated ones from the constructor). */
    private transient MIndexAlgorithm[] algorithms;
    
    /** A simple lock for periodic algorithm serialization. */
    protected transient Lock serializingLock = new ReentrantLock();
    

    /**
     * Creates a new multi-algorithm overlay for the given array of algorithms.
     * @param algorithms the algorithms on which the operations are processed
     */
    public MultipleMIndexAlgorithm(Collection<? extends Algorithm> algorithms) {
        super(algorithms);
        setAlgorithms(algorithms);
    }

    /**
     * Creates a new multi-algorithm overlay for the given collection of algorithms.
     * @param algorithms the algorithms on which the operations are processed
     */
    @Algorithm.AlgorithmConstructor(description = "Constructor with created algorithms", arguments = {"array of running algorithms", "t/f if operation should be clonned before running"})
    public MultipleMIndexAlgorithm(Algorithm[] algorithms) {
        this(Arrays.asList(algorithms));
    }

    /**
     * Initializes the internal array of encapsulated algorithms with the given collection.
     * @param algorithms the algorithms for the array
     */
    private void setAlgorithms(Collection<? extends Algorithm> algorithms) {
        this.algorithms = algorithms.toArray(new MIndexAlgorithm[algorithms.size()]);
    }

    /** Deserialize and set the transient fields. */
    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        setAlgorithms(getAlgorithms());
        serializingLock = new ReentrantLock();        
    }
//
//    @Override
//    public void setOperationsThreadPool(ExecutorService operationsThreadPool) {
//        super.setOperationsThreadPool(operationsThreadPool); //To change body of generated methods, choose Tools | Templates.
//        for (Algorithm algorithm : algorithms) {
//            if (algorithm.getOperationsThreadPool() == null) {
//                algorithm.setOperationsThreadPool(operationsThreadPool);
//            }
//        }
//    }
//



    /**
     * Checks all the M-Index algorithms, if their dynamic cell tree was not modified and
     *  serializes the whole algorithm if it was changed. THis method is typically called
     *  periodically.
     * @param serializationFile file to serialize the whole algorithm into
     * @throws IOException if the serialization failed
     */
    public void checkModifiedAndStore(String serializationFile) throws IOException {
        serializingLock.lock();
        try {
            boolean anyModified = false;
            for (MIndexAlgorithm mIndexAlgorithm : algorithms) {
                anyModified = anyModified || mIndexAlgorithm.getmIndex().isDynamicTreeModified();
            }
            if (anyModified) {
                // Store algorithm to file
                this.storeToFile(serializationFile);
            }
        } finally {
            serializingLock.unlock();
        }
    }

    @Override
    protected void beforeStoreToFile(String filepath) {
        super.beforeStoreToFile(filepath);
        for (MIndexAlgorithm mIndexAlgorithm : algorithms) {
            mIndexAlgorithm.beforeStoreToFile(filepath);
        }
    }

    @Override
    protected void afterStoreToFile(String filepath, boolean successful) {
        super.afterStoreToFile(filepath, successful); 
        for (MIndexAlgorithm mIndexAlgorithm : algorithms) {
            mIndexAlgorithm.afterStoreToFile(filepath, successful);
        }
    }        
    
    /**
     * Sets the period of the automatic temporary-closeable checking of the buckets.
     * @param period the new checking period in milliseconds;
     *          zero value means disable the checking
     */
    public void autoCloseBuckets(long period) {
        for (MIndexAlgorithm mIndexAlgorithm : algorithms) {
            mIndexAlgorithm.autoCloseBuckets(period);
        }
    }

    public Collection<PPPCalculator> getAllPPCalculators() {
        ArrayList<PPPCalculator> pPCalculators = new ArrayList<>(algorithms.length);
        for (MIndexAlgorithm alg : algorithms) {
            pPCalculators.add(alg.getPPCalculator());
        }
        return pPCalculators;
    }
    
    protected VoronoiTreeIndex getFirstMIndex() {
        return algorithms[0].getmIndex();
    }
    
    /** 
     * Clears (removes) all the data stored in all referenced M-Index algorithms. The structure of the 
     *  dynamic cell trees remains untouched.
     */
    public void clearData() throws AlgorithmMethodException {
        for (MIndexAlgorithm mIndexAlgorithm : algorithms) {
            mIndexAlgorithm.clearData();
        }
    }

    @Override
    protected void registerProcessors() {
        super.registerProcessors();
        // this is a fallback for all distance-based single-query search operations
        this.handler.register(ProcessorConfiguration.create(KNNOperationApprox.class, RankedAnswer.class)
                .withProcessor(operation -> new QueryOperationProcessor(operation, KNNOperationApprox.class, getAlgorithms()))
                .executeParallel(threadPool).build());
        this.handler.register(ProcessorConfiguration.create(RangeOperationApprox.class, RankedAnswer.class)
                .withProcessor(operation -> new QueryOperationProcessor(operation, RangeOperationApprox.class, getAlgorithms()))
                .executeParallel(threadPool).build());
        this.handler.register(ProcessorConfiguration.create(KNNRangeOperationApprox.class, RankedAnswer.class)
                .withProcessor(operation -> new QueryOperationProcessor(operation, KNNRangeOperationApprox.class, getAlgorithms()))
                .executeParallel(threadPool).build());

    }


    class QueryOperationProcessor extends AlgorithmsOperationProcessor<QueryObjectOperation, RankedAnswer> {

        public QueryOperationProcessor(QueryObjectOperation operation, Class<? extends QueryObjectOperation> subClass,
                                       Collection<Algorithm> algorithms) {
            super(getPartialOperation(operation, subClass, getAlgorithmsCount()), getMergingProcessor(operation), algorithms);
        }

    }

    protected static AnswerProvider<RankedAnswer> getMergingProcessor(QueryObjectOperation operation) {
        final RankedSortedCollection answerCollection = RankedAnswer.Builder.getAnswerCollection(operation);
        answerCollection.setIgnoringDuplicates(true);
        return new RankedAnswer.Builder<>(operation, answerCollection, RankedAnswer.class);
    }

    protected static QueryObjectOperation getPartialOperation(QueryObjectOperation operation, Class<? extends QueryObjectOperation> subClass, int algorithmCount) {
        if (operation instanceof ApproximateSearch) {
            final OperationBuilder<? extends QueryObjectOperation> subOpbuilder = OperationBuilder.create(subClass);

            // put all params from the original operation (especially, query object, "k" and/or radius)
            subOpbuilder.addParams(operation.getParams().getMap());

            // ...and rewrite the approximate settings
            subOpbuilder.addParam(ApproximateSearch.APPROX_PARAM_FIELD, ((ApproximateSearch) operation).getApproxParam() / algorithmCount);

            // ...and rewrite the query object to contain the filter
            final IndexObject queryObject = new IndexObject(operation.getQueryObject());
            final Set<Object> duplicateDCSetFinal = Collections.newSetFromMap(new ConcurrentHashMap<>());
            queryObject.setIndexField(DuplicateRefineFilter.FILTER_STRING, new DuplicateRefineFilter(duplicateDCSetFinal));
            subOpbuilder.addParam(QueryObject.QUERY_OBJECT_FIELD, queryObject);

            return subOpbuilder.build();
        }
        return operation;
    }

    @Override
    public String toString() {
        StringBuilder stringBuilder = new StringBuilder(super.toString());
        int i = 1;
        for (MIndexAlgorithm mIndexAlgorithm : algorithms) {
            stringBuilder.append("Sub algorithm ").append(i++).append(": \n").append(mIndexAlgorithm.toString());
        }
        return stringBuilder.toString();
    }

}
