/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.algorithm;


import messif.algorithm.Algorithm;
import messif.algorithm.AlgorithmMethodException;
import messif.operation.answer.AbstractAnswer;
import messif.operation.crud.DeleteOperation;
import messif.operation.crud.InsertOperation;
import messif.operation.processing.ProcessorConfiguration;
import messif.operation.processing.impl.OneStepProcessor;
import messif.operation.search.GetRandomObjectsOperation;
import messif.operation.answer.ListingAnswer;
import messif.utility.ExtendedProperties;
import mindex.MetricIndexes;
import mindex.PPPCalculator;
import mindex.VoronoiTreeIndex;
import mindex.privates.AnalyzeAvgCellDistanceOperation;
import mindex.processor.DeleteOperationProcessor;
import mindex.processor.GetRandomObjectsProcessor;
import mindex.processor.InsertOperationProcessor;

import java.io.IOException;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * This is the main M-Index centralized algorithm.
 *
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public abstract class VoronoiAlgorithm extends Algorithm {

    /** Class serial id for serialization. */
    private static final long serialVersionUID = 203203L;

    /** This object encapsulates the M-Index logic, config, storage, navigation. */
    protected final VoronoiTreeIndex mIndex;

    /** A simple lock for periodic algorithm serialization. */
    protected transient Lock serializingLock = new ReentrantLock();

    /**
     * Creates the algorithm given name and already created MIndex config instance.
     * @param name alg. name
     * @param mIndex created MIndex config object
     */
    protected VoronoiAlgorithm(String name, VoronoiTreeIndex mIndex, ExtendedProperties configuration) {
        super(name, configuration);
        this.mIndex = mIndex;
        MetricIndexes.logger.info(mIndex.toString());
    }

    @Override
    public synchronized boolean init() {
        final boolean retVal = super.init();
        if (retVal)
            serializingLock = new ReentrantLock();
        return retVal;
    }

    protected void registerProcessors() {
        super.registerProcessors();
        this.handler.register(ProcessorConfiguration.create(InsertOperation.class, ListingAnswer.class)
                .withProcessor((operation) -> new InsertOperationProcessor(operation, mIndex))
                .executeParallel(threadPool).build());
        this.handler.register(ProcessorConfiguration.create(DeleteOperation.class, ListingAnswer.class)
                .withProcessor((operation) -> new DeleteOperationProcessor(operation, mIndex))
                .executeParallel(threadPool).build());
        this.handler.register(ProcessorConfiguration.create(GetRandomObjectsOperation.class, ListingAnswer.class)
                .withProcessor((operation) -> new GetRandomObjectsProcessor(operation, mIndex))
                .executeParallel(threadPool).build());

        this.handler.register(ProcessorConfiguration.create(AnalyzeAvgCellDistanceOperation.class, AbstractAnswer.class)
                .withProcessor((operation) -> new AuxiliaryOperationProcessor(operation))
                .executeSequential().build());

    }

    /**
     * Stores the whole algorithm into the passed file if the structure of
     *  the dynamic cell tree was changed. This method is typically called periodically (every 1s).
     * @param serializationFile file to store the algorithm to
     * @throws IOException
     */
    public void checkModifiedAndStore(String serializationFile) throws IOException {
        serializingLock.lock();
        try {
            if (mIndex.isDynamicTreeModified()) {
                // Store algorithm to file
                this.storeToFile(serializationFile);
            }
        } finally {
            serializingLock.unlock();
        }
    }
    
    @Override
    public void beforeStoreToFile(String filepath) {
        super.beforeStoreToFile(filepath);
        mIndex.setDynamicTreeModified(false);
    }

    @Override
    public void afterStoreToFile(String filepath, boolean successful) {
        super.afterStoreToFile(filepath, successful); 
        if (!successful) {
            mIndex.setDynamicTreeModified(true);
        }
    }

    // ************************      Getters and setters      ************************* //

    /** 
     * Clears (removes) all the data stored in this M-Index. The structure of the 
     *  dynamic cell tree remains untouched.
     */
    public void clearData() throws AlgorithmMethodException {
        mIndex.clearData();
    }

    public PPPCalculator getPPCalculator() {
        return mIndex.getPPCalculator();
    }


    // *************************     Operation processing     *************************** //    

    protected class AuxiliaryOperationProcessor extends OneStepProcessor<AuxiliaryOperation, AbstractAnswer> {

        AbstractAnswer answer;

        public AuxiliaryOperationProcessor(AuxiliaryOperation operation) {
            super(operation);
        }

        @Override
        protected void process() throws Exception {
            this.answer = getOperation().process(mIndex);
        }

        @Override
        public AbstractAnswer finish() {
            return answer;
        }
    }

    /**
     * Prints detailed information about this M-Index - including the dynamic cluster tree
     *  and numbers of objects stored in individual clusters.
     * WARNING: opens all disk buckets (files)
     * @return string representation of this M-Index
     */
    public String toRichString() {
        StringBuilder strBuf = new StringBuilder(super.toString()).append(":\n").append(mIndex.toRichString());
        return strBuf.toString();
    }

    /**
     * Prints info about this M-Index.
     * @return string representation of this M-Index
     */
    @Override
    public String toString() {
        StringBuilder strBuf = new StringBuilder(super.toString()).append(":\n").append(mIndex.toString());
        return strBuf.toString();
    }
}
