/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.navigation;


import messif.algorithm.AlgorithmMethodException;
import messif.operation.search.RankingOperation;
import messif.operation.answer.RankedAnswer;
import mindex.VoronoiTreeIndex;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * Abstract class encapsulates a leaf node of a cluster tree for a dynamic M-Index.
 *
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 * @param <O> type of data stored in the Voronoi leaf cell
 */
public abstract class VoronoiLeafCell<O> extends VoronoiCell<O> {

    /** Class id for serialization. */
    private static final long serialVersionUID = 412001L;
    
    /** Number of objects stored in this leaf node. */
    protected volatile int objectCount = 0;
    //protected AtomicInteger objectCount = new AtomicInteger();

    /** ReadWrite lock of this leaf node. */
    protected transient ReadWriteLock readWriteLock = new ReentrantReadWriteLock(true);

    /** If false this LeafCell was split and should not be used anymore. */
    private volatile VoronoiInternalCell<O> subtitutionOfThisLeaf = null;
    
    /**
     * Constructor given all parameters mandatory for the parent class. The storage and the covered interval are set to null.
     * @param voronoiIndex M-Index logic
     * @param parentNode parent node of this node
     */
    public VoronoiLeafCell(VoronoiTreeIndex voronoiIndex, VoronoiInternalCell parentNode) {
        super(voronoiIndex, parentNode);
    }

    /**
     * Copy constructor.
     * @param copyFrom cell to copy the fields from (no clonning).
     */
    public VoronoiLeafCell(VoronoiLeafCell copyFrom) {
        super(copyFrom);
        this.objectCount = copyFrom.objectCount;
        this.subtitutionOfThisLeaf = copyFrom.subtitutionOfThisLeaf;
    }
    
    /**
     * Read the serialized disk storage from an object stream.
     * @param in the object stream from which to read the disk storage
     * @throws IOException if there was an I/O error during deserialization
     * @throws ClassNotFoundException if there was an unknown object in the stream
     */
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        // Proceed with standard deserialization first
        in.defaultReadObject();
        readWriteLock = new ReentrantReadWriteLock(true);
    }    
    
    // ***********************      Abstract method implementation   ********************************* //
    
    @Override
    public int getObjectCount() {
        return objectCount;
    }

    @Override
    public int getLevel() {
        return parentNode.getLevel() + 1;
    }

    @Override
    public VoronoiLeafCell getResponsibleLeaf(short[] ppp, boolean createPath) throws AlgorithmMethodException {
        return this;
    }

    @Override
    public short[] getPppForReading() {
        return getParentNode().getChildPPP(this);
    }

    @Override
    public short[] getPppForReading(short thisIndex) {
        short[] retVal = Arrays.copyOf(parentNode.getPppForReading(), getLevel());
        retVal[retVal.length - 1] = thisIndex;
        return retVal;
    }
    
    // ***************************      Data and storage manipulation    ********************* //
        
    /**
     * Tries to insert all the objects into the bucket managed by this LeafCell.
     * If {@code CapacityFullException} is thrown during insertion a list of
     * objects that were not inserted if returned
     * @param objects to be inserted to the bucket
     * @return request for split, if the leaf became full during the insert
     * @throws AlgorithmMethodException 
     */
    public abstract SplitConfiguration insertObjects(List<O> objects) throws AlgorithmMethodException;
        
    /**
     * Method that deletes given objects from corresponding Bucket. If an instance
     * of MergeConfiguration is returned, the upper layer should merge this LeafCell (TODO).
     * @param objects to be deleted
     * @param deleteLimit maximal number of objects data-equal to given object that will be deleted from this node
     * @param checkLocator is true, the objects to be deleted are not located based on data but by locator; if locator
     *  of given object is <b>null</b>, the locator is ignored!!!
     * @return objects that were actually deleted
     * @throws AlgorithmMethodException 
     */
    public abstract Collection<O> deleteObjects(List<O> objects, int deleteLimit, boolean checkLocator) throws AlgorithmMethodException;

    
    /**
     * This method should be called only when the {@code insertObjects} method returns non-empty list of objects
     * (= {@code CapacityFullException} was throw during insertion of objects into bucket managed by this LeafCell)
     * or when the {@code checkAgileMultibucketPartitioning} returns true
     * @param splitConfiguration settings for the split operation
     * @return true, if the split was successfully accomplished
     * @throws AlgorithmMethodException if anything goes wrong
     */
    public abstract boolean split(SplitConfiguration splitConfiguration) throws AlgorithmMethodException;

    /**
     * Fallback to process given query operation on all data.
     * @param queryProcessor queryProcessor evaluating the operation to be processed on this bucket
     * @throws AlgorithmMethodException if anything goes wrong
     */
    public abstract void processQuery(RankingOperation.Helper<? extends RankingOperation, ? extends RankedAnswer> queryProcessor) throws AlgorithmMethodException;

    /**
     * Return one random object stored in this leaf cell, or null if the cell is empty.
     * @return one random object stored in this leaf cell, or null if the cell is empty
     */
    public abstract O getRandomObject();

    // *******************************   Auxiliary  methods   **************************** //
    
    public final void readLock() {
        readWriteLock.readLock().lock();
    }

    public final void readUnlock() {
        readWriteLock.readLock().unlock();
    }

    public final void writeLock() {
        readWriteLock.writeLock().lock();
    }

    public void writeUnlock() {
        readWriteLock.writeLock().unlock();
    }

    /**
     * Checks if this LeafCell was already split
     * @return true if this LeafCell is valid
     *         false if this LeafCell was split
     */
    public boolean isValid() {
        return subtitutionOfThisLeaf == null;
    }

    /**
     * If this cell is valid, this method returns null; otherwise, it returns
     * the new node that substituted this node in the Voronoi cell tree
     * structure.
     * @return the new node that substituted this node in the Voronoi cell tree; null, if this node is valid
     */
    public VoronoiInternalCell<O> getSubtitutionOfThisLeaf() {
        return subtitutionOfThisLeaf;
    }

    /**
     * If this leaf cell was split, then this method should be called to properly set the internal node that substitutes
     *  this node in the PPP-Tree.
     * @param newInternalSubstitute newly created internal cell
     */
    protected void invalidateThisNode(VoronoiInternalCell<O> newInternalSubstitute) {
        this.subtitutionOfThisLeaf = newInternalSubstitute;
    }
    
    @Override
    public String toString() {
        return new StringBuffer(": ").append( isValid() ).append(" bucket # ").append(" storing ").append(getObjectCount()).append("\n").toString();
    }

}
