/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processor;


import messif.algorithm.AlgorithmMethodException;
import messif.bucket.Bucket;
import messif.bucket.LocalBucket;
import messif.bucket.index.SearchDataObjectIterator;
import messif.operation.params.ApproximateSearch;
import messif.operation.search.QueryObjectOperation;
import messif.operation.answer.RankedAnswer;
import messif.statistic.OperationStatistics;
import mindex.MIndex;
import mindex.MetricIndexes;
import mindex.VoronoiTreeIndex;
import mindex.navigation.MIndexLeafCell;
import mindex.navigation.MIndexPreciseLeafCell;
import mindex.navigation.VoronoiCell;
import mindex.navigation.VoronoiLeafCell;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.logging.Level;

/**
 * Navigation processor for kNN or range query evaluated in approximate manner on M-Index.
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class ApproxQueryProcessor extends ApproxQueryProcessorBase {

    /** Limit of the search costs to be spent by this context (and sub-contexts). */
    @Nonnull
    protected final ApproxSearchCosts predictedCosts;

    /** Statistics about the real costs spent. */
    protected ApproxSearchCosts realCosts;

    /** Flag saying whether the key intervals should be checked for this approximate operation */
    protected final boolean checkKeyInterval;

    /** Flag saying whether the double-pivot filtering should be applied */
    protected final boolean checkDoubleFiltering;

    /** Thread-safe cache, used when multi-buckets used and the flag "bucket.process_whole_multi_bucket" is true */
    protected MultiBucketCache cache;

    /** Set of accessed buckets to be checked when multi-buckets are used */
    protected final Set<Bucket> processedBuckets;

    protected final QueryObjectOperation.Helper helper;

    /**
     * Creates new approximate navigation processor for given operation and M-Index. The priority queue is initialized.
     * @param operation approximate operation - typically kNN or range (should be the same for approximate range)
     * @param approxSettings the approximate settings for the operation (typically it can be the same instance as operation)
     * @param mIndex M-Index object with config and dynamic cell tree
     * @throws AlgorithmMethodException if the M-Index cannot determine the PP for query object
     */
    public ApproxQueryProcessor(QueryObjectOperation operation, ApproximateSearch approxSettings, MIndex mIndex)  {
        //this(operation, ApproxSearchCosts.getMIndexApproxSearchCosts(approxSettings, mIndex), mIndex, new SortedCollection<>(pPPComparator), true);
        this(operation, ApproxSearchCosts.getMIndexApproxSearchCosts(approxSettings, mIndex), mIndex, new PriorityQueue(1000, pPPComparator), true);
    }

    protected ApproxQueryProcessor(QueryObjectOperation operation, ApproxSearchCosts approxCosts, VoronoiTreeIndex mIndex, Queue<SinglePPPRankable> queue, boolean checkDoubleFiltering)  {
        super(operation, mIndex, queue);
        checkKeyInterval = mIndex.isApproxCheckKeyInterval() && mIndex.isPreciseSearch();
        this.checkDoubleFiltering = checkDoubleFiltering;
        cache = (mIndex.isMultiBuckets() && ! mIndex.isApproxProcessWholeMultiBucket()) ? new MultiBucketCache() : null;
        processedBuckets = (mIndex.isMultiBuckets() && mIndex.isApproxProcessWholeMultiBucket()) ? new HashSet() : null;
        predictedCosts = approxCosts;
        if (predictedCosts == null) {
            throw new IllegalArgumentException("the approximate settings are NULL and should not be");
        }

        this.helper = new QueryObjectOperation.Helper(operation, mIndex.getDistanceFunc());
    }

    @Override
    protected void initProcessor() throws AlgorithmMethodException {
        super.initProcessor();
        try {
            predictedCosts.updateSingleCosts(ApproximateSearch.Type.ABS_DC_COUNT, (int) OperationStatistics.getOpStatisticCounter("DistanceComputations").get());
            realCosts = predictedCosts.clone();
        } catch (CloneNotSupportedException ex) {
            MetricIndexes.logger.log(Level.SEVERE, null, ex);
            throw new AlgorithmMethodException(ex);
        }
    }

    @Override
    public RankedAnswer finish() {
        if (realCosts != null) {
            realCosts.setThreadStatistics();
        }
        if (cache != null) {
            OperationStatistics.getOpStatisticCounter("RedundantReadObjects").set(cache.getCurrentObjectCount());
            OperationStatistics.getOpStatisticCounter("RedundantReadCells").set(cache.getCurrentCellCount());
        }
        return helper.getAnswer();
    }

    @Override
    protected boolean stopTheProcessing(boolean isAsynchronous) {
        return ! (isAsynchronous ? predictedCosts : realCosts).checkCostCondition();
    }

    @Override
    protected boolean filterTheCell(RankableCell bestCell) {
        if (!checkDoubleFiltering)
            return false;
        return filterByDoublePivot(bestCell.cell, bestCell.getPPP(), helper.getAnswerThreshold());
    }

    @Override
    protected boolean alreadyProcessed(VoronoiLeafCell bestLeaf) {
        predictedCosts.updatePredictMaxCosts(bestLeaf);
        // if the leaf storage was already processed, ignore it
        return  (processedBuckets != null && processedBuckets.contains(((MIndexLeafCell) bestLeaf).getStorage()));
    }
    
    @Override
    protected void processItem(SinglePPPRankable processingItem) throws AlgorithmMethodException {
        try {
            // evaluate the operation on relevant data
            long currentDCCount = OperationStatistics.getOpStatisticCounter("DistanceComputations").get();
            if (! (((RankableCell) processingItem).cell instanceof MIndexLeafCell)) {
                throw new AlgorithmMethodException("this approx. operation can be processed only on MIndexLeafCellData");
            }
            MIndexLeafCell leafCell = (MIndexLeafCell) ((RankableCell) processingItem).cell;
            int accessedObjects = processOnLeaf(leafCell, ((RankableCell) processingItem).getPPP());

            if (realCosts != null) {
                realCosts.updateAfterSearch(accessedObjects, (int) (OperationStatistics.getOpStatisticCounter("DistanceComputations").get() - currentDCCount));
            }
            //return operation;
        } finally {
            ((MIndexLeafCell) ((RankableCell) processingItem).cell).readUnlock();
        }
    }
    
    /**
     * Process the approximate query on the data from given M-Index Voronoi leaf cell 
     *  (call standard {@link messif.operation.search.QueryObjectOperation.Helper#evaluate(Iterator)}} on relevant data).
     * @param leaf M-Index Voronoi leaf cell
     * @return number of objects processed (accessed) by the operation
     * @throws AlgorithmMethodException if the evaluate method fails
     */
    protected int processOnLeaf(MIndexLeafCell leaf, short [] leafPPP) throws AlgorithmMethodException {
        if (leaf.getStorage() == null) {
            return 0;
        }
        float radius = helper.getAnswerThreshold();

        // if there are multibuckets and whole multibuckets are always processed
        if (processedBuckets != null) {
            synchronized (processedBuckets) {
                // if this bucket was already fully processed
                if (! processedBuckets.add(leaf.getStorage())) {
                    return 0;
                }
            }
        } 
        
        // try to filter the leaf out by multiple application of Double-pivot constraint (and Pivot-Range constraint)
        if (filterOut(leaf, leafPPP, radius)) {
            return 0;
        }

        if (processedBuckets != null) {
            leaf.processQuery(helper);
            OperationStatistics.getOpStatisticCounter("AccessedBuckets").add();
            return ((LocalBucket) leaf.getStorage()).getObjectCount();
        }
        
        //  if whole buckets are always processed
        if (! checkKeyInterval && cache == null || !(leaf.getStorage() instanceof LocalBucket)) {
            leaf.processQuery(helper);
            OperationStatistics.getOpStatisticCounter("AccessedBuckets").add();
            return leaf.getObjectCount();
        }

        // use the interval            
        if (checkKeyInterval) {
            SearchDataObjectIterator iterator = RangeNavigationProcessor.getIntervalIterator(leaf,
                    RangeNavigationProcessor.createRelevantInterval((MIndexPreciseLeafCell) leaf, radius, queryPPPDistCalculator.getQueryPivotDistances()), cache, mIndex);
            helper.evaluate(iterator);
            return iterator.getCount();
        }
        // for pure disk buckets ignore the limits completely
        helper.evaluate(cache.processBucket(leaf.getStorage(), leaf).iterator());
        return leaf.getObjectCount();
    }

    /**
     * This method checks whether the specified cluster can be filtered out using repetitive application of double-pivot distance constraint.
     * @param node M-Index Voronoi cell
     * @param radius actual query radius
     * @return true only if given node can be filtered out according to given query and radius
     */
    protected boolean filterByDoublePivot(VoronoiCell node, short [] nodePPP, float radius) {
        if (!checkDoubleFiltering)
            return false;
        return MetricIndexes.filterOutCluster(queryPPPDistCalculator.getQueryPivotDistances(), queryPPPDistCalculator.getQueryPPP(), nodePPP, radius);
    }

    /**
     * This method tries to filter out the leaf cell by repetitive application of double-pivot constraint {@link #filterByDoublePivot(VoronoiCell, short[], float)}
     *   and by range-pivot constraint, if the interval is available.
     * @param leaf leaf Voronoi cell to filter out
     * @param radius actual query radius
     * @return true, if given cell can be filtered out by the two mentioned constraints
     */
    protected boolean filterOut(MIndexLeafCell leaf, short [] leafPPP, float radius) {
        // if radius of the query is already set, try to filter the bucket out using multiple double-pivot constraint
        if (filterByDoublePivot(leaf, leafPPP, radius)) {
            return true;
        }
        // try to filter using range-pivot distance constraint
        if (leaf instanceof MIndexPreciseLeafCell) {
            if (! ((MIndexPreciseLeafCell) leaf).getCoveredInterval().intersect(RangeNavigationProcessor.createRelevantInterval((MIndexPreciseLeafCell) leaf, radius, queryPPPDistCalculator.getQueryPivotDistances()))) {
                return true;
            }
        }
        return false;
    }

}
