/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processor;


import messif.algorithm.AlgorithmMethodException;
import messif.indexing.IndexObject;
import messif.operation.processing.impl.QueueOperationProcessor;
import messif.operation.search.QueryObjectOperation;
import messif.operation.answer.RankedAnswer;
import mindex.MetricIndexes;
import mindex.VoronoiTreeIndex;
import mindex.distance.PartialQueryPPPDistanceCalculator;
import mindex.distance.QueryPPPDistanceCalculator;
import mindex.navigation.MinQueryDistanceRevisor;
import mindex.navigation.VoronoiCell;
import mindex.navigation.VoronoiInternalCell;
import mindex.navigation.VoronoiLeafCell;

import java.util.Comparator;
import java.util.Iterator;
import java.util.Map;
import java.util.Queue;

/**
 * Navigation processor for kNN or range query evaluated in approximate manner on M-Index.
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public abstract class ApproxQueryProcessorBase extends QueueOperationProcessor<QueryObjectOperation, RankedAnswer, SinglePPPRankable> {

    /** Calculator of "distance" between the query and PPP */
    protected final QueryPPPDistanceCalculator queryPPPDistCalculator;

    /** Counters of objects that were added to the priority queue */
    public volatile int nodesAddedToQ;

    /** Number of items processed from the queue */
    public volatile int queueItemsProcessed;

    /** The Voronoi tree object */
    protected final VoronoiTreeIndex mIndex;

    protected ApproxQueryProcessorBase(QueryObjectOperation operation, VoronoiTreeIndex mIndex, Queue<SinglePPPRankable> queue)  {
        super(operation, queue, false);
        this.mIndex = mIndex;
        queryPPPDistCalculator = MetricIndexes.getQueryPPPDistCalculator(operation, mIndex);
    }

    @Override
    protected void initProcessor() throws AlgorithmMethodException {
        super.initProcessor();
        processingItems.add(new RankableCell(mIndex.getVoronoiCellTree(), queryPPPDistCalculator));
        nodesAddedToQ = 1;
    }

    /**
     * This method finds the next leaf cell in the priority queue and returns it. Filtering is applied
     * @param isAsynchronous true, if the processing using this processor is currently asynchronous
     * @return the next {@code RankableCell} to be processed, or null if the top item is NOT instance of {@link RankableCell} (or if queue is empty)
     */
    @Override
    protected synchronized RankableCell getNextProcessingItem(boolean isAsynchronous) {
        // check the stop condition, if to be checked;
        //if ((isAsynchronous ? predictedCosts : realCosts) != null && (!(isAsynchronous ? predictedCosts : realCosts).checkCostCondition())) {
        if (stopTheProcessing(isAsynchronous)) {
            processingItems.clear();
            return null;
        }

        SinglePPPRankable head;
        while ((head = processingItems.peek()) != null && (head instanceof RankableCell)) {
            RankableCell bestCell = (RankableCell) processingItems.poll();
            if (bestCell.reviseDistance(queryPPPDistCalculator)) {
                processingItems.add(bestCell);
                continue;
            }
            queueItemsProcessed ++;
            //if (checkDoubleFiltering && filterByDoublePivot(bestCell.cell, bestCell.getPPP(), helper.getAnswerThreshold())) {
            if (filterTheCell(bestCell)) {
                continue;
            }
            VoronoiLeafCell bestLeaf;
            if (bestCell.cell instanceof VoronoiLeafCell || bestCell.cell.getLevel() >= queryPPPDistCalculator.getMaxLevel()) {
//                if (bestCell.cell.getObjectCount() <= 0) {
//                    continue;
//                }
                ((VoronoiLeafCell) bestCell.cell).readLock();
                if (bestCell.cell instanceof VoronoiLeafCell) {
                    bestLeaf = (VoronoiLeafCell) bestCell.cell;
                    // check whether the leaf node wasn't split in the meanwhile
                    if (! bestLeaf.isValid()) {
                        processingItems.add(new RankableCell(bestLeaf.getSubtitutionOfThisLeaf(), bestCell.rankingDistance, bestCell.thisIndex));
                        nodesAddedToQ++;
                        ((VoronoiLeafCell) bestCell.cell).readUnlock();
                        continue;
                    }
                    if (alreadyProcessed(bestLeaf)) {
                        ((VoronoiLeafCell) bestCell.cell).readUnlock();
                        continue;
                    }
                }
                return bestCell;
            }
            // generate the contexts for all subnodes
            for (Iterator<Map.Entry<Short, VoronoiCell<IndexObject>>> childNodes = ((VoronoiInternalCell) bestCell.cell).getChildNodes(); childNodes.hasNext(); ) {
                Map.Entry<Short, VoronoiCell<IndexObject>> entry = childNodes.next();
                processingItems.add(new RankableCell(entry.getValue(), queryPPPDistCalculator, bestCell, entry.getKey()));
                nodesAddedToQ ++;
            }
        }
        return null;
    }

    protected abstract boolean stopTheProcessing(boolean isAsynchronous);

    protected abstract boolean filterTheCell(RankableCell bestCell);

    protected abstract boolean alreadyProcessed(VoronoiLeafCell bestLeaf);

//
//    @Override
//    public RankedAnswer finish() {
//        if (realCosts != null) {
//            realCosts.setThreadStatistics();
//        }
//        if (cache != null) {
//            OperationStatistics.getOpStatisticCounter("RedundantReadObjects").set(cache.getCurrentObjectCount());
//            OperationStatistics.getOpStatisticCounter("RedundantReadCells").set(cache.getCurrentCellCount());
//        }
//        return helper.getAnswer();
//    }
//
    /**
     * Internal class used to compare approx context to order contexts increasingly.
     */
    protected static class RankableCell implements SinglePPPRankable {
        
        private static final float EQUALITY_TOLERANCE = 0.0001f;
        
        /** M-Index penalty of the cluster - distance from current query object */
        private float rankingDistance;

        /** M-Index penalty of the cluster - distance from current query object */
        private final float ownQueryDistance;
        
        /** M-Index cluster */
        protected final VoronoiCell cell;
        
        /** This cells last PP index (useful for leaf cells) */
        protected final short thisIndex;
        
        /** Flag saying if the distance can be revised or not */
        protected boolean wasRevised = false;

        @Override
        public float getQueryDistance() {
            return rankingDistance;
        }
        
        public VoronoiCell getCell() {
            return cell;
        }

        public final short[] getPPP() {
            return cell.getPppForReading(thisIndex);
        }

        public float getOwnQueryDistance() {
            return ownQueryDistance;
        }
        
        /** Sets the cell and calculates approximate penalty */
        public RankableCell(final VoronoiCell cell, QueryPPPDistanceCalculator calculator, RankableCell topCell, short thisIndex) {
            this.cell = cell;
            this.thisIndex = thisIndex;
            if (calculator instanceof PartialQueryPPPDistanceCalculator) {
                this.rankingDistance = topCell.ownQueryDistance + ((PartialQueryPPPDistanceCalculator) calculator).getPartialQueryDistance(thisIndex, topCell.cell.getLevel() + 1);
            } else {
                this.rankingDistance = calculator.getQueryDistance(getPPP());
            }
            this.ownQueryDistance = this.rankingDistance;
        }

        public RankableCell(final VoronoiCell rootCell, QueryPPPDistanceCalculator calculator) {
            this.cell = rootCell;
            this.ownQueryDistance = calculator.getQueryDistance(cell.getPppForReading());
            this.rankingDistance = ownQueryDistance;
            this.thisIndex = (short) 0;
        }

        public RankableCell(final VoronoiCell cell, float distance, short thisIndex) {
            this.cell = cell;
            this.ownQueryDistance = distance;
            this.rankingDistance = ownQueryDistance;
            this.thisIndex = thisIndex;
        }
        
        protected boolean reviseDistance(QueryPPPDistanceCalculator calculator) {
            try {
                if (!(calculator instanceof PartialQueryPPPDistanceCalculator && cell instanceof MinQueryDistanceRevisor && !wasRevised)) {
                    return false;
                }
                this.rankingDistance = ((MinQueryDistanceRevisor) cell).reviseMinDistance(((PartialQueryPPPDistanceCalculator) calculator), ownQueryDistance);
                if (ownQueryDistance > rankingDistance + EQUALITY_TOLERANCE) {
                    MetricIndexes.logger.warning("original distance was larger than the revised: " + ownQueryDistance + ", " + rankingDistance + ", " + cell.getPppForReading());
                }
                return (ownQueryDistance < rankingDistance);
            } finally {
                wasRevised = true;
            }
        }
        
    };
    
    /**
     * Comparator between objects or M-Index cells according to query-PPP distance.
     */
    public static Comparator<SinglePPPRankable> pPPComparator = (o1, o2) -> Float.compare(o1.getQueryDistance(), o2.getQueryDistance());
    
}
