/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processor;


import messif.algorithm.AlgorithmMethodException;
import messif.bucket.Bucket;
import messif.bucket.LocalBucket;
import messif.bucket.index.SearchDataObjectIterator;
import messif.indexing.IndexObject;
import messif.operation.params.ApproximateSearch;
import messif.operation.processing.impl.QueueOperationProcessor;
import messif.operation.search.QueryObjectOperation;
import messif.operation.answer.RankedAnswer;
import messif.statistic.OperationStatistics;
import mindex.MIndex;
import mindex.MetricIndexes;
import mindex.VoronoiTreeIndex;
import mindex.distance.PartialQueryPPPDistanceCalculator;
import mindex.distance.QueryPPPDistanceCalculator;
import mindex.navigation.*;

import java.util.*;
import java.util.logging.Level;

/**
 * Navigation processor for kNN or range query evaluated in approximate manner on M-Index.
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
@Deprecated
public class ApproxQueryProcessorOld extends QueueOperationProcessor<QueryObjectOperation, RankedAnswer, SinglePPPRankable> {

    /** Queue of M-Index leaf cells sorted by M-Index approx. penalty */
    //protected final Queue<SinglePPPRankable> processingItems;

    /** Calculator of "distance" between the query and PPP */
    protected final QueryPPPDistanceCalculator queryPPPDistCalculator;

    /** Limit of the search costs to be spent by this context (and sub-contexts). */
    protected final ApproxSearchCosts predictedCosts;

    /** Statistics about the real costs spent. */
    protected ApproxSearchCosts realCosts;

    /** Flag saying whether the key intervals should be checked for this approximate operation */
    protected final boolean checkKeyInterval;

    /** Flag saying whether the double-pivot filtering should be applied */
    protected final boolean checkDoubleFiltering;

    /** Counters of objects that were added to the priority queue */
    public volatile int nodesAddedToQ;

    /** Number of items processed from the queue */
    public volatile int queueItemsProcessed;

    /** Thread-safe cache, used when multi-buckets used and the flag "bucket.process_whole_multi_bucket" is true */
    protected MultiBucketCache cache;

    /** Set of accessed buckets to be checked when multi-buckets are used */
    protected final Set<Bucket> processedBuckets;

    /** the M-Index object */
    protected final VoronoiTreeIndex mIndex;

    protected final QueryObjectOperation.Helper helper;

    /**
     * Creates new approximate navigation processor for given operation and M-Index. The priority queue is initialized.
     * @param operation approximate operation - typically kNN or range (should be the same for approximate range)
     * @param approxSettings the approximate settings for the operation (typically it can be the same instance as operation)
     * @param mIndex M-Index object with config and dynamic cell tree
     * @throws AlgorithmMethodException if the M-Index cannot determine the PP for query object
     */
    public ApproxQueryProcessorOld(QueryObjectOperation operation, ApproximateSearch approxSettings, MIndex mIndex)  {
        //this(operation, ApproxSearchCosts.getMIndexApproxSearchCosts(approxSettings, mIndex), mIndex, new SortedCollection<>(pPPComparator), true);
        this(operation, ApproxSearchCosts.getMIndexApproxSearchCosts(approxSettings, mIndex), mIndex, new PriorityQueue(1000, pPPComparator), true);
    }

    protected ApproxQueryProcessorOld(QueryObjectOperation operation, ApproxSearchCosts approxCosts, VoronoiTreeIndex mIndex, Queue<SinglePPPRankable> queue, boolean checkDoubleFiltering)  {
        super(operation, queue, false);
        this.mIndex = mIndex;
        queryPPPDistCalculator = MetricIndexes.getQueryPPPDistCalculator(operation, mIndex);
        checkKeyInterval = mIndex.isApproxCheckKeyInterval() && mIndex.isPreciseSearch();
        this.checkDoubleFiltering = checkDoubleFiltering;
        cache = (mIndex.isMultiBuckets() && ! mIndex.isApproxProcessWholeMultiBucket()) ? new MultiBucketCache() : null;
        processedBuckets = (mIndex.isMultiBuckets() && mIndex.isApproxProcessWholeMultiBucket()) ? new HashSet() : null;
        predictedCosts = approxCosts;

        this.helper = new QueryObjectOperation.Helper(operation, mIndex.getDistanceFunc());
    }

    @Override
    protected void initProcessor() throws AlgorithmMethodException {
        super.initProcessor();
        processingItems.add(new RankableCell(mIndex.getVoronoiCellTree(), queryPPPDistCalculator));
        nodesAddedToQ = 1;
        try {
            if (predictedCosts != null) {
                predictedCosts.updateSingleCosts(ApproximateSearch.Type.ABS_DC_COUNT, (int) OperationStatistics.getOpStatisticCounter("DistanceComputations").get());
                realCosts = predictedCosts.clone();
            } else {
                realCosts = null;
            }
        } catch (CloneNotSupportedException ex) {
            MetricIndexes.logger.log(Level.SEVERE, null, ex);
            throw new AlgorithmMethodException(ex);
        }
    }

    /**
     * This method finds the next leaf cell in the priority queue and returns it. Filtering is applied
     * @param isAsynchronous true, if the processing using this processor is currently asynchronous
     * @return the next {@code RankableCell} to be processed, or null if the top item is NOT instance of {@link RankableCell} (or if queue is empty)
     */
    @Override
    protected synchronized RankableCell getNextProcessingItem(boolean isAsynchronous) {
        // check the stop condition, if to be checked;
        if ((isAsynchronous ? predictedCosts : realCosts) != null && (!(isAsynchronous ? predictedCosts : realCosts).checkCostCondition())) {
            processingItems.clear();
            return null;
        }

        SinglePPPRankable head;
        while ((head = processingItems.peek()) != null && (head instanceof RankableCell)) {
            RankableCell bestCell = (RankableCell) processingItems.poll();
            if (bestCell.reviseDistance(queryPPPDistCalculator)) {
                processingItems.add(bestCell);
                continue;
            }
            queueItemsProcessed ++;
            if (checkDoubleFiltering && filterByDoublePivot(bestCell.cell, bestCell.getPPP(), helper.getAnswerThreshold())) {
                continue;
            }
            VoronoiLeafCell bestLeaf;
            if (bestCell.cell instanceof VoronoiLeafCell || bestCell.cell.getLevel() >= queryPPPDistCalculator.getMaxLevel()) {
//                if (bestCell.cell.getObjectCount() <= 0) {
//                    continue;
//                }
                ((VoronoiLeafCell) bestCell.cell).readLock();
                if (bestCell.cell instanceof VoronoiLeafCell) {
                    bestLeaf = (VoronoiLeafCell) bestCell.cell;
                    // check whether the leaf node wasn't split in the meanwhile
                    if (! bestLeaf.isValid()) {
                        processingItems.add(new RankableCell(bestLeaf.getSubtitutionOfThisLeaf(), bestCell.rankingDistance, bestCell.thisIndex));
                        nodesAddedToQ++;
                        ((VoronoiLeafCell) bestCell.cell).readUnlock();
                        continue;
                    }
                    if (predictedCosts != null) {
                        predictedCosts.updatePredictMaxCosts(bestLeaf);
                    }
                    // if the leaf storage was already processed, ignore it
                    if (processedBuckets != null && processedBuckets.contains(((MIndexLeafCell) bestLeaf).getStorage())) {
                        ((VoronoiLeafCell) bestCell.cell).readUnlock();
                        continue;
                    }
                }
                return bestCell;
            }
            // generate the contexts for all subnodes
            for (Iterator<Map.Entry<Short, VoronoiCell<IndexObject>>> childNodes = ((VoronoiInternalCell) bestCell.cell).getChildNodes(); childNodes.hasNext(); ) {
                Map.Entry<Short, VoronoiCell<IndexObject>> entry = childNodes.next();
                processingItems.add(new RankableCell(entry.getValue(), queryPPPDistCalculator, bestCell, entry.getKey()));
                nodesAddedToQ ++;
            }
        }
        return null;
    }

    @Override
    protected void processItem(SinglePPPRankable processingItem) throws AlgorithmMethodException {
        try {
            // evaluate the operation on relevant data
            long currentDCCount = OperationStatistics.getOpStatisticCounter("DistanceComputations").get();
            if (! (((RankableCell) processingItem).cell instanceof MIndexLeafCell)) {
                throw new AlgorithmMethodException("this approx. operation can be processed only on MIndexLeafCellData");
            }
            MIndexLeafCell leafCell = (MIndexLeafCell) ((RankableCell) processingItem).cell;
            int accessedObjects = processOnLeaf(leafCell, ((RankableCell) processingItem).getPPP());

            if (realCosts != null) {
                realCosts.updateAfterSearch(accessedObjects, (int) (OperationStatistics.getOpStatisticCounter("DistanceComputations").get() - currentDCCount));
            }
            //return operation;
        } finally {
            ((VoronoiLeafCell) ((RankableCell) processingItem).cell).readUnlock();
        }
    }

    /**
     * Process the approximate query on the data from given M-Index Voronoi leaf cell
     *  (call standard {@link QueryObjectOperation.Helper#evaluate(Iterator)}} on relevant data).
     * @param leaf M-Index Voronoi leaf cell
     * @return number of objects processed (accessed) by the operation
     * @throws AlgorithmMethodException if the evaluate method fails
     */
    protected int processOnLeaf(MIndexLeafCell leaf, short [] leafPPP) throws AlgorithmMethodException {
        if (leaf.getStorage() == null) {
            return 0;
        }
        float radius = helper.getAnswerThreshold();

        // if there are multibuckets and whole multibuckets are always processed
        if (processedBuckets != null) {
            synchronized (processedBuckets) {
                // if this bucket was already fully processed
                if (! processedBuckets.add(leaf.getStorage())) {
                    return 0;
                }
            }
        } 
        
        // try to filter the leaf out by multiple application of Double-pivot constraint (and Pivot-Range constraint)
        if (filterOut(leaf, leafPPP, radius)) {
            return 0;
        }

        if (processedBuckets != null) {
            leaf.processQuery(helper);
            OperationStatistics.getOpStatisticCounter("AccessedBuckets").add();
            return ((LocalBucket) leaf.getStorage()).getObjectCount();
        }
        
        //  if whole buckets are always processed
        if (! checkKeyInterval && cache == null || !(leaf.getStorage() instanceof LocalBucket)) {
            leaf.processQuery(helper);
            OperationStatistics.getOpStatisticCounter("AccessedBuckets").add();
            return leaf.getObjectCount();
        }

        // use the interval            
        if (checkKeyInterval) {
            SearchDataObjectIterator iterator = RangeNavigationProcessor.getIntervalIterator(leaf,
                    RangeNavigationProcessor.createRelevantInterval((MIndexPreciseLeafCell) leaf, radius, queryPPPDistCalculator.getQueryPivotDistances()), cache, mIndex);
            helper.evaluate(iterator);
            return iterator.getCount();
        }
        // for pure disk buckets ignore the limits completely
        helper.evaluate(cache.processBucket(leaf.getStorage(), leaf).iterator());
        return leaf.getObjectCount();
    }

    @Override
    public RankedAnswer finish() {
        if (realCosts != null) {
            realCosts.setThreadStatistics();
        }
        if (cache != null) {
            OperationStatistics.getOpStatisticCounter("RedundantReadObjects").set(cache.getCurrentObjectCount());
            OperationStatistics.getOpStatisticCounter("RedundantReadCells").set(cache.getCurrentCellCount());
        }
        return helper.getAnswer();
    }
    
    /**
     * This method checks whether the specified cluster can be filtered out using repetitive application of double-pivot distance constraint.
     * @param node M-Index Voronoi cell
     * @param radius actual query radius
     * @return true only if given node can be filtered out according to given query and radius
     */
    protected boolean filterByDoublePivot(VoronoiCell node, short [] nodePPP, float radius) {
        return MetricIndexes.filterOutCluster(queryPPPDistCalculator.getQueryPivotDistances(), queryPPPDistCalculator.getQueryPPP(), nodePPP, radius);
    }

    /**
     * This method tries to filter out the leaf cell by repetitive application of double-pivot constraint {@link #filterByDoublePivot(VoronoiCell, short[], float)}
     *   and by range-pivot constraint, if the interval is available.
     * @param leaf leaf Voronoi cell to filter out
     * @param radius actual query radius
     * @return true, if given cell can be filtered out by the two mentioned constraints
     */
    protected boolean filterOut(MIndexLeafCell leaf, short [] leafPPP, float radius) {
        // if radius of the query is already set, try to filter the bucket out using multiple double-pivot constraint
        if (filterByDoublePivot(leaf, leafPPP, radius)) {
            return true;
        }
        // try to filter using range-pivot distance constraint
        if (leaf instanceof MIndexPreciseLeafCell) {
            if (! ((MIndexPreciseLeafCell) leaf).getCoveredInterval().intersect(RangeNavigationProcessor.createRelevantInterval((MIndexPreciseLeafCell) leaf, radius, queryPPPDistCalculator.getQueryPivotDistances()))) {
                return true;
            }
        }
        return false;
    }

    /**
     * Internal class used to compare approx context to order contexts increasingly.
     */
    protected static class RankableCell implements SinglePPPRankable {
        
        private static final float EQUALITY_TOLERANCE = 0.0001f;
        
        /** M-Index penalty of the cluster - distance from current query object */
        private float rankingDistance;

        /** M-Index penalty of the cluster - distance from current query object */
        private final float ownQueryDistance;
        
        /** M-Index cluster */
        protected final VoronoiCell cell;
        
        /** This cells last PP index (useful for leaf cells) */
        protected final short thisIndex;
        
        /** Flag saying if the distance can be revised or not */
        protected boolean wasRevised = false;

        @Override
        public float getQueryDistance() {
            return rankingDistance;
        }
        
        public VoronoiCell getCell() {
            return cell;
        }

        public final short[] getPPP() {
            return cell.getPppForReading(thisIndex);
        }

        public float getOwnQueryDistance() {
            return ownQueryDistance;
        }
        
        /** Sets the cell and calculates approximate penalty */
        public RankableCell(final VoronoiCell cell, QueryPPPDistanceCalculator calculator, RankableCell topCell, short thisIndex) {
            this.cell = cell;
            this.thisIndex = thisIndex;
            if (calculator instanceof PartialQueryPPPDistanceCalculator) {
                this.rankingDistance = topCell.ownQueryDistance + ((PartialQueryPPPDistanceCalculator) calculator).getPartialQueryDistance(thisIndex, topCell.cell.getLevel() + 1);
            } else {
                this.rankingDistance = calculator.getQueryDistance(getPPP());
            }
            this.ownQueryDistance = this.rankingDistance;
        }

        public RankableCell(final VoronoiCell rootCell, QueryPPPDistanceCalculator calculator) {
            this.cell = rootCell;
            this.ownQueryDistance = calculator.getQueryDistance(cell.getPppForReading());
            this.rankingDistance = ownQueryDistance;
            this.thisIndex = (short) 0;
        }

        public RankableCell(final VoronoiCell cell, float distance, short thisIndex) {
            this.cell = cell;
            this.ownQueryDistance = distance;
            this.rankingDistance = ownQueryDistance;
            this.thisIndex = thisIndex;
        }
        
        protected boolean reviseDistance(QueryPPPDistanceCalculator calculator) {
            try {
                if (!(calculator instanceof PartialQueryPPPDistanceCalculator && cell instanceof MinQueryDistanceRevisor && !wasRevised)) {
                    return false;
                }
                this.rankingDistance = ((MinQueryDistanceRevisor) cell).reviseMinDistance(((PartialQueryPPPDistanceCalculator) calculator), ownQueryDistance);
                if (ownQueryDistance > rankingDistance + EQUALITY_TOLERANCE) {
                    MetricIndexes.logger.warning("original distance was larger than the revised: " + ownQueryDistance + ", " + rankingDistance + ", " + cell.getPppForReading());
                }
                return (ownQueryDistance < rankingDistance);
            } finally {
                wasRevised = true;
            }
        }
        
    };
    
    /**
     * Comparator between objects or M-Index cells according to query-PPP distance.
     */
    public static Comparator<SinglePPPRankable> pPPComparator = (o1, o2) -> Float.compare(o1.getQueryDistance(), o2.getQueryDistance());
    
}
