/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processor;


import messif.algorithm.AlgorithmMethodException;
import messif.operation.crud.CRUDOperation;
import messif.operation.processing.impl.QueueOperationProcessor;
import messif.operation.answer.ListingAnswer;
import mindex.VoronoiTreeIndex;
import mindex.navigation.VoronoiLeafCell;

import java.util.*;

/**
 * Implementation of an operation processor interface that can handle all CRUD operations (insert, delete, update, and
 *  may be also {@link messif.operation.crud.GetObjectsOperation}.
 *
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public abstract class CRUDOperationProcessor<O> extends QueueOperationProcessor<CRUDOperation, ListingAnswer, Map.Entry<VoronoiLeafCell<O>, List<O>>> {

    /** VoronoiTreeIndex (NavigationDirectory) into which the objects should be inserted */
    protected final VoronoiTreeIndex<O> mIndex;

    /** If true, the Voronoi tree path is created, if necessary (this is probably "insert" operation). */
    protected final boolean createPath;

    final CRUDOperation.Helper helper;


    protected CRUDOperationProcessor(boolean createPath, CRUDOperation operation, VoronoiTreeIndex<O> mIndex, CRUDOperation.Helper helper) {
        super(operation);
        this.mIndex = mIndex;
        this.createPath = createPath;
        this.helper = helper;
    }

    /**
     * Computes MIndexKey for each of the objects and ands it to the list of its responsible leaf cell.
     * If allowed - initializing the objects (computing distances to pivots) is done in parallel.
     * @throws AlgorithmMethodException if anything goes wrong
     */
    @Override
    protected void initProcessor() throws AlgorithmMethodException {
        Map<VoronoiLeafCell<O>, List<O>> cellObjectMap = new HashMap<>();
        for (O object : mIndex.initObjects(operation.getObjects())) {
            addToCellObjectMap(cellObjectMap, object);
        }
        addProcessingItems(cellObjectMap.entrySet());
    }

    protected void addObjectsToQeue(Collection<O> initializedObjects) throws AlgorithmMethodException {
        Map<VoronoiLeafCell<O>, List<O>> cellObjectMap = new HashMap<>();
        for (O object : initializedObjects) {
            addToCellObjectMap(cellObjectMap, object);
        }
        addProcessingItems(cellObjectMap.entrySet());
    }

    /**
     * For the given INITIALIZED object (i.e. {@code VoronoiTreeIndex.getPPP(object)} != null) find the responsible leaf
     *   cell add it to the global cell object map and if not already locked, acquire the read lock on the leaf cell
     * @param object to be added to the list for its responsible MIndexLeafCell
     * @throws AlgorithmMethodException
     */
    private void addToCellObjectMap(Map<VoronoiLeafCell<O>, List<O>> cellObjectMap, O object) throws AlgorithmMethodException {
        VoronoiLeafCell leaf = mIndex.getVoronoiCellTree().getResponsibleLeaf(mIndex.getPPP(object), createPath);
        if (leaf == null)
            return;
        List<O> objects = cellObjectMap.get(leaf);
        if (objects == null) {
            cellObjectMap.put(leaf, objects = new ArrayList<>());
        }
        objects.add(object);
    }

    @Override
    public ListingAnswer finish() {
        return helper.getAnswer();
    }

}
