/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processor;


import messif.algorithm.AlgorithmMethodException;
import messif.operation.crud.DeleteOperation;
import messif.operation.crud.answer.DeleteAnswer;
import mindex.VoronoiTreeIndex;
import mindex.navigation.VoronoiLeafCell;

import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * Implementation of NavigationProcessor interface that can handle DeleteOperation
 * and is thread safe - the object is deleted only once even if more threads
 * call processNext simultaneously 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class DeleteOperationProcessor<O> extends CRUDOperationProcessor<O> {

    /**
     * Create the delete processor from a delete operation and current M-Index settings.
     * @param operation
     * @param mIndex
     * @throws AlgorithmMethodException
     */
    public DeleteOperationProcessor(DeleteOperation operation, VoronoiTreeIndex mIndex) {
        super(false, operation, mIndex, new DeleteOperation.Helper(operation));
    }

    @Override
    protected void processItem(Map.Entry<VoronoiLeafCell<O>, List<O>> processingItem) throws AlgorithmMethodException {
        final VoronoiLeafCell<O> leaf = processingItem.getKey();
        leaf.writeLock();
        try {
            if (!leaf.isValid()) {
                addObjectsToQeue(processingItem.getValue());
                return;
            }
            final Collection<O> deletedObjects = leaf.deleteObjects(processingItem.getValue(), 0, true);
            for (O deletedObject : deletedObjects) {
                helper.answerBuilder.add(mIndex.getDataObject(deletedObject), DeleteAnswer.OBJECT_DELETED);
            }
        } finally {
            leaf.writeUnlock();
        }
    }

}
