/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processor;

import messif.operation.processing.AsyncOperationProcessor;
import messif.operation.search.GetRandomObjectsOperation;
import messif.operation.answer.ListingAnswer;
import mindex.VoronoiTreeIndex;
import mindex.navigation.VoronoiCell;
import mindex.navigation.VoronoiInternalCell;
import mindex.navigation.VoronoiLeafCell;

import java.util.*;
import java.util.concurrent.Callable;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Processor for the {@link GetRandomObjectsOperation}. For each randomly selected object, it traverses
 *  the PPP-Tree in a random way and from the selected leaf, it returns a random object.
 * @author xnovak8
 */
public class GetRandomObjectsProcessor<O> implements AsyncOperationProcessor<GetRandomObjectsOperation, ListingAnswer> {

    /** Operation to process. */
    private final GetRandomObjectsOperation operation;
    
    /** Root of the PPP-Tree. */
    private final VoronoiInternalCell<O> voronoiTreeRoot;

    protected final Random random = new Random(System.currentTimeMillis());

    /** Counter of the threads - each is selecting one object. */
    private final AtomicInteger counter;

    /** Answer builder. */
    protected final ListingAnswer.Builder answerBuilder;

    /** The VoronoiIndex helper. */
    protected final VoronoiTreeIndex<O> mIndex;

    /**
     * Creates new process given an operation and PPP-Tree root. 
     * @param operation operation to process (defines number of objects to be returned)
     * @param mIndex the Voronoi tree index
     */
    public GetRandomObjectsProcessor(GetRandomObjectsOperation operation, VoronoiTreeIndex<O> mIndex) {
        this.operation = operation;
        this.voronoiTreeRoot = mIndex.getVoronoiCellTree();
        this.counter = new AtomicInteger(operation.getObjectCount());
        this.answerBuilder = new ListingAnswer.Builder(operation, ListingAnswer.class);
        this.mIndex = mIndex;
    }    
    
    /**
     * This methods selects an object randomly from given leaf node.
     * @param leaf PPP-Tree leaf node to select an object from
     * @return randomly selected object
     */
    protected O getOneRandomObject(VoronoiLeafCell<O> leaf) {
        leaf.readLock();
        try {
            return leaf.getRandomObject();
        } finally {
            leaf.readUnlock();
        }
    }
    
    /**
     * This method selects an object from the given internal node (it calls the method recursively).
     * @param internal internal node of the PPP-Tree
     * @return selected object (from a leaf node)
     */
    protected O getOneRandomObject(VoronoiInternalCell<O> internal) {
        Iterator<Map.Entry<Short, VoronoiCell<O>>> childIt = internal.getChildNodes();
        List<VoronoiCell<O>> childNodes = new ArrayList<>();
        while (childIt.hasNext()) {
            childNodes.add(childIt.next().getValue());
        }
        if (childNodes.isEmpty()) {
            return null;
        }
        
        O retVal = null;
        do {
            VoronoiCell<O> nextNode = childNodes.get(random.nextInt(childNodes.size()));
            if (nextNode instanceof VoronoiLeafCell) {
                retVal = getOneRandomObject((VoronoiLeafCell<O>) nextNode);
            }
            if (nextNode instanceof VoronoiInternalCell) {
                retVal = getOneRandomObject((VoronoiInternalCell<O>) nextNode);
            }
            childNodes.remove(nextNode);
        } while (retVal == null && (! childNodes.isEmpty()));
        
        return retVal;
    }
    
    @Override
    public Callable<Void> processStepAsynchronously() throws InterruptedException {
        if (counter.getAndDecrement() <= 0) {
            return null;
        }
        return () -> {
            addToAnswer(getOneRandomObject(voronoiTreeRoot));
            return null;
        };
    }

    @Override
    public GetRandomObjectsOperation getOperation() {
        return operation;
    }

    @Override
    public boolean processStep() throws InterruptedException {
        if (counter.getAndDecrement() <= 0) {
            return false;
        }
        addToAnswer(getOneRandomObject(voronoiTreeRoot));
        return true;
    }

    @Override
    public ListingAnswer finish() {
        return answerBuilder.getAnswer();
    }

    /**
     * Simply add a next randomly selected object to the query answer. This method can be override.
     * @param nextRandom object to be inserted to the query answer
     */
    protected void addToAnswer(O nextRandom) throws IllegalArgumentException {
        if (nextRandom != null) {
            answerBuilder.add(mIndex.getDataObject(nextRandom));
        }
    }    


}
