/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processor;

import messif.algorithm.AlgorithmMethodException;
import messif.operation.crud.InsertOperation;
import messif.operation.crud.answer.InsertAnswer;
import mindex.VoronoiTreeIndex;
import mindex.navigation.SplitConfiguration;
import mindex.navigation.VoronoiLeafCell;

import java.util.List;
import java.util.Map;

/**
 * Implementation of NavigationProcessor interface that can handle both
 * InsertOperation and BulkInsertOperation.
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class InsertOperationProcessor<O> extends CRUDOperationProcessor<O> {

    /**
     * Inits the processor for a multi-object bulk-insert operation. The objects are initialized
     *  by M-Index (set M-Index pivot permutation) and added to the cell map.
     * @param operation single-object insert operation
     * @param mIndex M-Index object
     */
    public InsertOperationProcessor(InsertOperation operation, VoronoiTreeIndex mIndex) {
        super(true, operation, mIndex, new InsertOperation.Helper(operation));
    }

    /**
     * Insert given objects into the given leaf cell. If some of the objects
     *   could NOT be inserted (i.e. bucket is full) split the leaf cell.
     * @param processingItem a map entry with the leaf cell and list of objects to be inserted
     * @throws AlgorithmMethodException if the initialization goes wrong
     */
    @Override
    protected void processItem(Map.Entry<VoronoiLeafCell<O>, List<O>> processingItem) throws AlgorithmMethodException {
        final VoronoiLeafCell<O> leaf = processingItem.getKey();
        leaf.writeLock();
        SplitConfiguration<O> splitConfig = null;
        try {
            if (!leaf.isValid()) {
                addObjectsToQeue(processingItem.getValue());
                return;
            }
            splitConfig = leaf.insertObjects(processingItem.getValue());
            List<O> insertedObjects;
            if (splitConfig != null) { // if any splitConfig is returned, split the leaf
                // some of the objects weren't inserted into the leaf cell's bucket
                // so split the leaf cell, insert them into appropriate leaf cells.
//                leaf.getParentNode().splitLock();
//                try {
                leaf.split(splitConfig);
//                } finally {
//                    leaf.getParentNode().splitUnlock();
//                }

                // upper layers calling this processor need to handle proper insertion of objects after split by calling processStep() method
                addObjectsToQeue(splitConfig.getToReinsert());
                insertedObjects = splitConfig.getInsertedObjects();
            } else {
                insertedObjects = processingItem.getValue();
            }
            for (O insertedObject : insertedObjects) {
                helper.answerBuilder.add(mIndex.getDataObject(insertedObject), InsertAnswer.OBJECT_INSERTED);
            }
        } finally {
            leaf.writeUnlock();
        }
    }

}
