/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processor;

import messif.algorithm.AlgorithmMethodException;
import messif.bucket.Bucket;
import messif.bucket.BucketStorageException;
import messif.bucket.index.IndexComparator;
import messif.bucket.index.Search;
import messif.bucket.index.impl.AbstractSearch;
import messif.data.DataObject;
import messif.data.util.DataObjectIterator;
import messif.data.util.DataObjectList;
import messif.statistic.OperationStatistics;
import messif.utility.Interval;
import mindex.MIndexKeys;
import mindex.navigation.MIndexInternalCell;
import mindex.navigation.MIndexLeafCell;
import mindex.navigation.VoronoiLeafCell;

import java.util.*;


/**
 * Temporary memory storage of read objects (aggregated by pivot permutation) which
 *  is used during query evaluation especially when multi-buckets are used and
 *  some data may be already read.
 * The cache is thread-safe.
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
*/
public class MultiBucketCache {
    
    /** Cache storage */
    protected Map<VoronoiLeafCell, List<DataObject>> cache = new HashMap<>();

    /** Visited buckets list */
    protected Set<Integer> visitedBuckets = new HashSet<>();

    private void putCluster(VoronoiLeafCell leafCell, List<DataObject> bucket) throws AlgorithmMethodException {
        if (cache.containsKey(leafCell)) {
            throw new AlgorithmMethodException("this cluster is already in the cache: " + leafCell);
        }
        cache.put(leafCell, bucket);
    }    
    
    public synchronized List<DataObject> processBucket(Bucket bucket, MIndexLeafCell leafCell) throws AlgorithmMethodException {
        List<DataObject> retVal;
        if (visitedBuckets.contains(bucket.getBucketID())) {
            retVal = cache.remove(leafCell);
            if (retVal == null) {
                throw new AlgorithmMethodException("leaf cell '" + leafCell.toString() + "' was empty but is to be processed locally");
            }
            return retVal;
        }
        
        visitedBuckets.add(bucket.getBucketID());
        int leafCellLevel = leafCell.getLevel();
        OperationStatistics.getOpStatisticCounter("AccessedBuckets").add();
        // find out, if this bucket is really a multibucket
        Map<Short,MIndexLeafCell> bucketNodes = ((MIndexInternalCell) leafCell.getParentNode()).getAllBucketNodes(bucket);
        if (bucketNodes.size() == 1) {
            retVal = new DataObjectList(bucket.getAllObjects());
        } else {
            for (VoronoiLeafCell structureLeafNode : bucketNodes.values()) {
                if (structureLeafNode.getObjectCount() > 0) {
                    putCluster(structureLeafNode, new ArrayList<>());
                }
            }
            for (DataObjectIterator it = bucket.getAllObjects(); it.hasNext(); ) {
                DataObject obj = it.next();
                cache.get(bucketNodes.get((MIndexKeys.readPivotIndex(obj, leafCellLevel)))).add(obj);
                //cache.get(MetricIndexes.readMIndexPP(obj).getClusterNumber(cellID.getLevel())).add(obj);
            }
            retVal  = cache.remove(leafCell);
            if (retVal == null) {
                throw new AlgorithmMethodException("cluster with key '" + leafCell.toString() + "' is empty but is to be processed locally");
            }
        }
        return retVal;
    }
    
    public Search<DataObject> processBucketWithInterval(Bucket bucket, MIndexLeafCell cellID, IndexComparator<Float, DataObject> comparator, Interval<Float> interval) throws AlgorithmMethodException {
        return new ObjectListSearch(processBucket(bucket, cellID), comparator, interval.getFrom(), interval.getTo());
    }
    
    public int getCurrentObjectCount() {
        int retVal = 0;
        for (List<DataObject> cluster : cache.values()) {
            retVal += cluster.size();
        }
        return retVal;
    }

    public int getCurrentCellCount() {
        return cache.size();
    }
    
    
    protected static class ObjectListSearch extends AbstractSearch<Float, DataObject> {

        protected final List<DataObject> objects;

        protected int currentIndex = -1;
        
        public ObjectListSearch(List<DataObject> objects, IndexComparator<Float, DataObject> comparator, Float fromKey, Float toKey) {
            super(comparator, fromKey, toKey);
            this.objects = objects;
        }
        
        @Override
        protected DataObject readNext() throws BucketStorageException {
            if (currentIndex < -1 || currentIndex >= objects.size() - 1) {
                return null;
            }
            return objects.get( ++ currentIndex );
        }

        @Override
        protected DataObject readPrevious() throws BucketStorageException {
            if (currentIndex < 1 || currentIndex >= objects.size() + 1) {
                return null;
            }
            return objects.get( -- currentIndex );
        }

        @Override
        public void close() {
        }
        
    }
    
}
