/*
 *  This file is part of M-Index library: https://bitbucket.org/disalab/m-index
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processor;


import messif.algorithm.AlgorithmMethodException;
import messif.bucket.LocalBucket;
import messif.bucket.index.Search;
import messif.bucket.index.SearchDataObjectIterator;
import messif.buckets.impl.FloatOrderedMemoryBucket;
import messif.indexing.IndexObject;
import messif.operation.processing.impl.QueueOperationProcessor;
import messif.operation.search.QueryObjectOperation;
import messif.operation.search.RangeOperation;
import messif.operation.answer.RankedAnswer;
import messif.statistic.OperationStatistics;
import messif.utility.Interval;
import mindex.MIndexKeys;
import mindex.VoronoiTreeIndex;
import mindex.navigation.*;

import java.util.Iterator;
import java.util.Map;

/**
 * Navigation processor for the precise range queries. It first determines those leaf cells that cannot be 
 *  pruned using repetitive application of the double-pivot constraint (or range pivot constraint). 
 *  Then, each of these leaves can be processed - using intervals within each leaf (object-pivot constraint).
 *
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class RangeNavigationProcessor extends QueueOperationProcessor<RangeOperation, RankedAnswer, VoronoiLeafCell> {

    /** The M-Index object */
    protected final VoronoiTreeIndex<IndexObject> mIndex;

    /** Distances between the query object and pivots */
    protected final float[] queryPivotDistances;
    
    /** Pivot-permutation indexes corresponding to given query */
    protected final short[] queryIndexes;
    
    /** Query radius (used often, therefore stored in this field) */
    protected final float queryRadius;
    
    /** Thread-safe cache, used when multi-buckets used */
    protected MultiBucketCache cache = null;

    /** Helper to build the range query answer. */
    protected final QueryObjectOperation.Helper helper;

    /**
     * Creates new approximate navigation processor for given operation and M-Index. The priority queue is initialized.
     * @param operation approximate kNN operation (should be the same for approximate range)
     * @param mIndex M-Index object with config and dynamic cell tree
     * @throws AlgorithmMethodException if the M-Index cannot determine the PP for query object
     */
    public RangeNavigationProcessor(RangeOperation operation, VoronoiTreeIndex<IndexObject> mIndex) throws InstantiationException {
        super(operation);
        if (! mIndex.isPreciseSearch()) {
            throw new InstantiationException("The range query can be processed only on precise M-Index");
        }
        this.mIndex = mIndex;
        this.helper = new QueryObjectOperation.Helper(operation, mIndex.getDistanceFunc());

        final IndexObject queryIndexObject = mIndex.initObject(operation.getQueryObject());
        queryPivotDistances = MIndexKeys.readObjPivotDists(queryIndexObject);
        queryIndexes = MIndexKeys.readMIndexPP(queryIndexObject);
        queryRadius = operation.getRadius();
        
        if (mIndex.isMultiBuckets()) {
            cache = new MultiBucketCache();
        }
    }

    @Override
    protected void initProcessor() throws AlgorithmMethodException {
        fillRelevantNodes((MIndexPreciseInternalCell) mIndex.getVoronoiCellTree());
        queueClose();
    }

    /**
     * Recursive method to traverse the M-Index dynamic cell tree in a breadth-first-manner
     *  and identify all leaf cell that cannot be filtered out using repetitive application
     *  of the double-pivot constraint.
     * @param treeRoot current internal node for tree traversal
     */
    private void fillRelevantNodes(MIndexPreciseInternalCell treeRoot) {
        int level = treeRoot.getLevel() + 1;
        //MIndexPP pivotCombination = treeRoot.                getPivotPermutation();
        short[] rootPPP = treeRoot.getPppForReading();
        // find the smallest pivot index, that was not used in higher cluster-tree levels
        int queryNextLevelToConsider = -1;
        boolean isActualFilteringIndexInCombination;
        do {
            queryNextLevelToConsider++;
            isActualFilteringIndexInCombination = false;
            //for (short i = 1; i <= level - 1; i++ ) {
            for (short i = 0; i < level - 1; i++ ) {
                //if (pivotCombination.readPivotIndex(i) == queryIndexes[queryNextLevelToConsider]) {
                if (rootPPP[i] == queryIndexes[queryNextLevelToConsider]) {
                    isActualFilteringIndexInCombination = true;
                    break;
                }
            }
        } while (isActualFilteringIndexInCombination);
        // iterate over all child nodes, try to filter them out and process recursively
        int i = 0;
        for (Iterator<Map.Entry<Short, VoronoiCell<IndexObject>>> iterator = treeRoot.getChildNodes(); iterator.hasNext(); i++) {
            VoronoiCell child = iterator.next().getValue();
            if (child == null) {
                continue;
            }
        
            if (queryPivotDistances[i] - queryPivotDistances[queryIndexes[queryNextLevelToConsider]] > 2f * queryRadius) {
                continue;
            }
            if (child instanceof MIndexPreciseInternalCell) {
                fillRelevantNodes((MIndexPreciseInternalCell) child);
            } else {
                processingItems.add((VoronoiLeafCell) child);
            }
        }
    }
    
    /**
     * Creates interval of keys relevant for given Voronoi cell and radius
     * @param leaf M-Index Voronoi leaf cell
     * @param radius actual query radius
     * @param queryPivotDistances array of distances between the query and the pivots
     *
     * @return interval of keys relevant for given Voronoi cell and radius
     */
    protected static Interval<Float> createRelevantInterval(MIndexPreciseLeafCell leaf, float radius, float [] queryPivotDistances) {
        float distance = queryPivotDistances[leaf.getPivotPermutation()[0]];

        // the interval uses the object-pivot distance constraint
        return new Interval<>(Math.max(distance - radius, 0f), distance + radius);
    }

    /**
     * Creates interval of keys relevant for given Voronoi cell and radius
     * @param leaf M-Index Voronoi leaf cell
     * @param radius actual query radius
     * @return interval of keys relevant for given Voronoi cell and radius
     */
    protected Interval<Float> createRelevantInterval(MIndexPreciseLeafCell leaf, float radius) {
        return createRelevantInterval(leaf, radius, queryPivotDistances);
    }

    /**
     * Given a leaf cell and M-Index key interval, this method creates an iterator over
     *  objects from this leaf whose keys are within this interval. If multi-buckets
     *  are used, the passed "cache" is used.
     * @param leaf M-Index leaf cell with data (with (multi-)bucket)
     * @param interval interval of M-Index keys generated by given query (range or kNN)
     * @param cache multi-bucket cache (can be null if multi-buckets are not used)
     * @param mIndex corresponding M-Index config
     * @return an iterator over objects from this leaf whose keys are within this interval
     * @throws AlgorithmMethodException if the processing goes wrong
     * @throws IllegalStateException if the underlying collection is modified in the meanwhile (should not happen)
     */
    protected static SearchDataObjectIterator getIntervalIterator(MIndexLeafCell leaf, Interval<Float> interval, MultiBucketCache cache, VoronoiTreeIndex mIndex) throws AlgorithmMethodException, IllegalStateException {
        Search objectsWithinInterval = (mIndex.isMultiBuckets()) ?
               cache.processBucketWithInterval(leaf.getStorage(), leaf, FloatOrderedMemoryBucket.createMIndexKeyComparator(leaf.getLevel()), interval) :
               ((LocalBucket) leaf.getStorage()).getIndex().search(FloatOrderedMemoryBucket.createMIndexKeyComparator(leaf.getLevel()), interval.getFrom(), interval.getTo());
        SearchDataObjectIterator iterator = new SearchDataObjectIterator(objectsWithinInterval, Integer.MAX_VALUE);

        if (! mIndex.isMultiBuckets()) {
            OperationStatistics.getOpStatisticCounter("AccessedBuckets").add();
        }
        return iterator;
    }

    @Override
    protected void processItem(VoronoiLeafCell processingItem) throws AlgorithmMethodException {
        // TODO: for remote buckets, the intervals should be used by the inteligent buckets in the other side;
        // TODO: for remote multi-buckets, the intervals should be grouped and then send
        MIndexPreciseLeafCell dataLeaf = (MIndexPreciseLeafCell) processingItem;
        if (dataLeaf.getStorage() == null) {
            return;
        }
        if (!(dataLeaf.getStorage() instanceof LocalBucket)) {
            dataLeaf.processQuery(helper);
            OperationStatistics.getOpStatisticCounter("AccessedBuckets").add();
        }
        
        Interval<Float> interval = createRelevantInterval(dataLeaf, queryRadius);
        // try to filter using range-pivot distance constraint
        if (dataLeaf.getCoveredInterval() != null && ! dataLeaf.getCoveredInterval().intersect(interval)) {
            return;
        }

        final SearchDataObjectIterator iterator = getIntervalIterator(dataLeaf, interval, cache, mIndex);

        // evaluate the operation on relevant data
        helper.evaluate(iterator);
    }

    @Override
    public RankedAnswer finish() {
        return helper.getAnswer();
    }
}